# pylint: disable=all
import numpy as np

def _constrain_to_multiple_of(x: int, multiple_of: int, min_val=0, max_val=None) -> int:
    x = np.array(x)
    y: np.ndarray = (np.round(x / multiple_of) * multiple_of).astype(int)
    if max_val is not None and y > max_val:
        y = (np.floor(x / multiple_of) * multiple_of).astype(int)
    if y < min_val:
        y = (np.ceil(x / multiple_of) * multiple_of).astype(int)
    return y.item()

def get_size(__height, __width, height, width, multiple_of) -> tuple[int, int]:
    # determine new height and width
    scale_height = __height / height
    scale_width = __width / width
    # keep aspect ratio
    if abs(1 - scale_width) < abs(1 - scale_height):
        # fit width
        scale_height = scale_width
    else:
        # fit height
        scale_width = scale_height
    new_height = _constrain_to_multiple_of(scale_height * height, multiple_of)
    new_width = _constrain_to_multiple_of(scale_width * width, multiple_of)

    return new_height, new_width



"""depth_representationpy -- module implementing a Depth Represenatation generic class"""
from overrides import overrides
import numpy as np
from vre.utils import MemoryData, DiskData, colorize_depth, ReprOut
from vre.representations import Representation, NpIORepresentation, NormedRepresentationMixin

class DepthRepresentation(Representation, NpIORepresentation, NormedRepresentationMixin):
    """DepthRepresentation. Implements depth task-specific stuff, like spectral map for plots."""
    def __init__(self, name: str, min_depth: float, max_depth: float, **kwargs):
        NormedRepresentationMixin.__init__(self)
        NpIORepresentation.__init__(self)
        Representation.__init__(self, name, **kwargs)
        self.min_depth = min_depth
        self.max_depth = max_depth

    @property
    @overrides
    def n_channels(self) -> int:
        return 1

    @overrides
    def disk_to_memory_fmt(self, disk_data: DiskData) -> MemoryData:
        return MemoryData(disk_data.clip(self.min_depth, self.max_depth))

    @overrides
    def make_images(self, data: ReprOut) -> np.ndarray:
        return (colorize_depth(data.output, percentiles=[1, 95]) * 255).astype(np.uint8)



# pylint: disable=all
import torch
import torch.nn.functional as F
from .util import rgb2hsv, normalize_channels, median_filter, like_an_image


def soft_seg(img, max_channels=3, use_median_filtering=False, as_image=False, internal_size=300):
    nb, _, ny, nx = img.shape
    img = normalize_channels(img)

    # rescale
    scale = internal_size / max(ny, nx)
    image_resized = F.interpolate(img, scale_factor=scale, mode='nearest', recompute_scale_factor=False)
    ry, rx = image_resized.shape[2:4]

    # pca data
    img_hsv = rgb2hsv(image_resized)
    map_sim = get_pca_data(img_hsv)

    # pca over patch maps
    map_sim = map_sim.view(nb, -1, ry * rx)
    mean_maps = map_sim.mean(dim=2, keepdim=True)
    map_sim -= mean_maps

    # eig
    val, eig = torch.linalg.eig(map_sim @ map_sim.transpose(2, 1))
    val, eig = val.real, eig.real   # matrix is symmetric

    # project maps on compressed space
    eig_valid = eig[:, :, :max_channels]
    seg = map_sim.transpose(2, 1) @ eig_valid
    seg = seg.view(nb, ry, rx, -1).permute(0, 3, 1, 2)

    # median filtering
    if use_median_filtering:
        seg = median_filter(seg, kernel_size=3)

    # chromatic adjustment to input image colors
    if as_image:
        seg = like_an_image(seg, image_resized, sampling_step=3)

    # resize output segmentations
    seg = F.interpolate(seg, size=(ny, nx), mode='bicubic', align_corners=False)
    seg = normalize_channels(seg, whole_block=True)

    return seg


def get_pca_data(img_hsv, dim_h=15, dim_s=11, dim_v=7, window=16, sample_step=30):
    nb, nc, ny, nx = img_hsv.shape
    dw = window // 2
    img_hsv = normalize_channels(img_hsv)

    # color mix
    h = (img_hsv[:, 0] * (dim_h - 1) + 1).round()
    s = (img_hsv[:, 1] * (dim_s - 1) + 1).round()
    v = (img_hsv[:, 2] * (dim_v - 1) + 1).round()
    img_col = h + (s - 1) * dim_h + (v - 1) * dim_h * dim_s - 1
    img_col = img_col.round().type(torch.long)

    # get reference patch locations
    xs, ys = torch.meshgrid(torch.arange(dw, nx - dw, dw), torch.arange(dw, ny - dw, dw), indexing="ij")
    ys, xs = ys.flatten(), xs.flatten()
    patch_colors = torch.zeros(nb, ys.numel(), dim_h * dim_v * dim_s, dtype=torch.bool, device=img_hsv.device)

    # patch has colors
    for b in range(nb):
        for i in range(ys.numel()):
            idx = img_col[b, ys[i] - dw:ys[i] + dw, xs[i] - dw:xs[i] + dw]
            patch_colors[b, i, idx.flatten()] = True

    # pca prep
    patch_colors = patch_colors.type(torch.float)
    mean_colors = patch_colors.mean(dim=1, keepdim=True)
    patch_colors -= mean_colors

    # eig
    val, eig = torch.linalg.eig(patch_colors.transpose(2, 1) @ patch_colors)
    val, eig = val.real, eig.real   # matrix is symmetric

    # get sampling patch locations
    xs, ys = torch.meshgrid(torch.arange(dw, nx - dw, sample_step), torch.arange(dw, ny - dw, sample_step), indexing="ij")
    ys, xs = ys.flatten(), xs.flatten()

    energy_sum = val.cumsum(dim=1)
    data = torch.zeros(nb, ys.numel(), ny, nx, device=img_hsv.device)
    for b in range(nb):
        # add up smallest values until 50% of total is reached
        max_vector = (energy_sum[b] > energy_sum[b, -1] * 0.5).nonzero()[0]
        eig_valid = eig[b, :, :max_vector + 1]

        for i in range(ys.numel()):
            # sample patch colors
            aux = img_col[b, ys[i] - dw:ys[i] + dw, xs[i] - dw:xs[i] + dw]
            aux_colors = torch.zeros(1, dim_h * dim_v * dim_s, device=img_hsv.device)
            aux_colors[0, aux.flatten()] = 1.
            aux_colors -= mean_colors[b]

            # how much of each color eigenvector is in this patch
            aux_match = (aux_colors @ eig_valid) @ eig_valid.T

            # vector mapped color values for this patch
            c0 = mean_colors[b] - aux_match
            c1 = mean_colors[b] + aux_match
            c0, c1 = c0.clamp(min=0), c1.clamp(min=0)

            # how much do colors in this patch appear across the image
            data[b, i] = c1[0, img_col[b]] / (c0[0, img_col[b]] + c1[0, img_col[b]] + torch.finfo(c0.dtype).eps)

    return data




"""cv2 utils. All the calls to opencv must be condensed here so eventually we can get rid of them"""
# pylint: disable=invalid-name, no-member
from pathlib import Path
import numpy as np
from ..logger import vre_logger as logger

try:
    import cv2
    cv2_CHAIN_APPROX_SIMPLE = cv2.CHAIN_APPROX_SIMPLE
    cv2_CHAIN_APPROX_NONE = cv2.CHAIN_APPROX_NONE
    cv2_RETR_TREE = cv2.RETR_TREE
    cv2_RETR_EXTERNAL = cv2.RETR_EXTERNAL
    cv2_RETR_CCOMP = cv2.RETR_CCOMP
    cv2_MORPH_OPEN = cv2.MORPH_OPEN
    cv2_MORPH_CLOSE = cv2.MORPH_CLOSE
    cv2_COLOR_RGB2BGR = cv2.COLOR_RGB2BGR
    cv2_COLOR_BGR2RGB = cv2.COLOR_BGR2RGB
    cv2_BORDER_CONSTANT = cv2.BORDER_CONSTANT
    cv2_LINE_AA = cv2.LINE_AA
    cv2_INTER_LINEAR = cv2.INTER_LINEAR
    cv2_INTER_NEAREST = cv2.INTER_NEAREST
except ImportError as e:
    logger.warning(f"Cannot import cv2: {e}")

def cv2_image_read(path: Path) -> np.ndarray:
    """wrapper on top of cv2.imread"""
    cv_res = cv2.imread(f"{path}")
    assert cv_res is not None, f"OpenCV returned None for '{path}'"
    bgr_image = cv_res[..., 0:3]
    b, g, r = cv2.split(bgr_image)
    image = cv2.merge([r, g, b]).astype(np.uint8)
    return image

def cv2_image_write(x: np.ndarray, path: Path):
    """wrapper on top of cv2.imwrite"""
    assert x.dtype == np.uint8, x.dtype
    res = cv2.imwrite(f"{path}", x[..., ::-1])
    assert res is not None and res is not False, f"Image {x.shape} could not be saved to '{path}'"

def cv2_image_resize(data: np.ndarray, height: int, width: int, interpolation: str, **kwargs) -> np.ndarray:
    """wraper on top of cv2.image_resize"""
    interpolation = {"nearest": cv2.INTER_NEAREST, "bilinear": cv2.INTER_LINEAR, "area": cv2.INTER_AREA,
                     "bicubic": cv2.INTER_CUBIC, "lanczos": cv2.INTER_LANCZOS4}[interpolation]
    _data = data if data.dtype == np.uint8 else data.astype(np.float32)
    res = cv2.resize(_data, dsize=(width, height), interpolation=interpolation, **kwargs).astype(data.dtype)
    res = res[..., None] if data.shape[-1] == 1 else res
    return res

def cv2_connectedComponents(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.connectedComponents(*args, **kwargs)

def cv2_Canny(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.Canny(*args, **kwargs)

def cv2_findContours(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.findContours(*args, **kwargs)

def cv2_boundingRect(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.boundingRect(*args, **kwargs)

def cv2_drawContours(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.drawContours(*args, **kwargs)

def cv2_morphologyEx(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.morphologyEx(*args, **kwargs)

def cv2_cvtColor(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.cvtColor(*args, **kwargs)

def cv2_connectedComponentsWithStats(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.connectedComponentsWithStats(*args, **kwargs)

def cv2_copyMakeBorder(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.copyMakeBorder(*args, **kwargs)

def cv2_getTextSize(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.getTextSize(*args, **kwargs)

def cv2_rectangle(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.rectangle(*args, **kwargs)

def cv2_putText(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.putText(*args, **kwargs)

def cv2_transform(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.transform(*args, **kwargs)

def cv2_warpAffine(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.warpAffine(*args, **kwargs)

def cv2_getRotationMatrix2D(*args, **kwargs):
    """Wrapper on top of cv2"""
    return cv2.getRotationMatrix2D(*args, **kwargs)



"""utils for vre"""
from typing import Any
from pathlib import Path
from datetime import datetime, timezone as tz
from collections import OrderedDict
from math import sqrt
from tqdm import tqdm
import numpy as np
import torch as tr
from vre.logger import vre_logger as logger

class DownloadProgressBar(tqdm):
    """requests + tqdm"""
    def update_to(self, b=1, bsize=1, tsize=None):
        """Callback from tqdm"""
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)

class FixedSizeOrderedDict(OrderedDict):
    """An OrderedDict with a fixed size. Useful for caching purposes."""
    def __init__(self, *args, maxlen: int = 0, **kwargs):
        self._maxlen = maxlen
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        OrderedDict.__setitem__(self, key, value)
        if self._maxlen > 0:
            if len(self) > self._maxlen:
                self.popitem(False)

def get_project_root() -> Path:
    """gets the root of this project"""
    return Path(__file__).parents[2].absolute()

def parsed_str_type(item: Any) -> str:
    """Given an object with a type of the format: <class 'A.B.C.D'>, parse it and return 'A.B.C.D'"""
    return str(type(item)).rsplit(".", maxsplit=1)[-1][0:-2]

def is_dir_empty(dir_path: Path, pattern: str = "*") -> bool:
    """returns true if directory is not empty, false if it is empty"""
    assert pattern.startswith("*"), pattern
    return len(list(dir_path.glob(pattern))) == 0

def is_git_lfs(path: Path) -> bool:
    """Returns true if a path is a git lfs link"""
    if path.is_dir():
        return False
    with open(path, "rb") as fp:
        try:
            return fp.read(7).decode("utf-8") == "version"
        except UnicodeDecodeError:
            return False

def get_closest_square(n: int) -> tuple[int, int]:
    """
    Given a stack of N images, find the closest square X>=N*N and return that.
    Note: There are only 2 rows possible between x^2 and (x+1)^2 because (x+1)^2 = x^2 + 2*x + 1, thus we can add two
    columns at most. If a 3rd column is needed, then closest lower bound is (x+1)^2 and we must use that.
    Example: 9: 3*3; 12 -> 3*3 -> 3*4 (3 rows). 65 -> 8*8 -> 8*9. 73 -> 8*8 -> 8*9 -> 9*9
    """
    x = int(sqrt(n))
    r, c = x, x
    c = c + 1 if c * r < n else c
    r = r + 1 if c * r < n else r
    assert (c + 1) * r > n and c * (r + 1) > n
    return r, c

def now_fmt() -> str:
    """Returns now() as a UTC isoformat string"""
    return datetime.now(tz=tz.utc).replace(tzinfo=None).isoformat(timespec="milliseconds")

def semantic_mapper(semantic_original: np.ndarray, mapping: dict[str, list[str]],
                    original_classes: list[str]) -> np.ndarray:
    """maps a bigger semantic segmentation to a smaller one"""
    assert len(semantic_original.shape) == 2, f"Only argmaxed data supported, got: {semantic_original.shape}"
    assert np.issubdtype(semantic_original.dtype, np.integer), semantic_original.dtype
    mapping_ix = {list(mapping.keys()).index(k): [original_classes.index(_v) for _v in v] for k, v in mapping.items()}
    flat_mapping = {}
    for k, v in mapping_ix.items():
        for _v in v:
            flat_mapping[_v] = k
    assert (A := len(set(flat_mapping))) == (B := len(set(original_classes))), (A, B)
    mapped_data = np.vectorize(flat_mapping.get)(semantic_original).astype(np.uint8)
    return mapped_data

def abs_path(x: str | Path) -> Path:
    """returns the absolute path of a string/path"""
    return Path(x).absolute()

def reorder_dict(data: dict[str, Any], keys: list[str]) -> dict[str, Any]:
    """simply puts in front the desired keys from the original dict, keeping the others intact"""
    assert (diff := set(keys).difference(data.keys())) == set(),diff
    for k in keys[::-1]:
        data = {k: data[k], **{k: v for k, v in data.items() if data != k}}
    return data

def str_maxk(s: str, k: int) -> str:
    """returns the string if it is smaller than k, otherwise returns the first k-2 and last 5"""
    return f"{s[0:k-7]}..{s[-5:]}" if len(s) > k else s

def array_blend(x: np.ndarray, y: np.ndarray, alpha: float | np.ndarray) -> np.ndarray:
    """Blends two arrays of the same shape with an alpha: number of array"""
    alpha_arr = np.asarray(alpha)
    assert x.shape == y.shape, (x.shape, y.shape)
    assert np.issubdtype(x.dtype, np.floating) and np.issubdtype(y.dtype, np.floating), (x.dtype, y.dtype)
    assert (alpha_arr >= 0).all() and (alpha_arr <= 1).all(), (alpha_arr.min(), alpha_arr.max())
    try:
        return (1 - alpha_arr) * x + alpha_arr * y # the actual blend :)
    except Exception as e:
        logger.info(f"Exception thrown: {e}.\nShapes: {x.shape=} {y.shape=} {alpha_arr.shape=}")
        raise e

def make_batches(frames: list[int], batch_size: int) -> list[int]:
    """return 1D array [start_frame, start_frame+bs, start_frame+2*bs... end_frame]"""
    if batch_size > len(frames):
        logger.warning(f"batch size {batch_size} is larger than #frames to process {len(frames)}.")
        batch_size = len(frames)
    batches, n_batches = [], len(frames) // batch_size + (len(frames) % batch_size > 0)
    for i in range(n_batches):
        batches.append(frames[i * batch_size: (i + 1) * batch_size])
    return batches

def vre_load_weights(path: Path) -> dict[str, tr.Tensor]:
    """load weights from disk. weights can be sharded as well. Sometimes we store this due to git-lfs big files bug."""
    logger.debug(f"Loading weights from '{path}'")

    if path.is_dir():
        res = {}
        for item in path.iterdir():
            res = {**res, **tr.load(item, map_location="cpu")}
        return res
    return tr.load(path, map_location="cpu")



"""DataStorer module -- Multi-threaded wrapper for DataWriter"""
from threading import Thread
from multiprocessing import cpu_count
from queue import Queue, Empty
from copy import deepcopy
from time import time

from .representations import ReprOut
from .data_writer import DataWriter
from .logger import vre_logger as logger

class DataStorer:
    """
    Equivalent of DataLoader on top of a Dataset -> DataStorer on top of a DataWriter
    Parameters:
    - data_writer The DataWriter object used by this DataStorer (akin to Dataset and DataLoader)
    - n_threads_data_storer The number of workers used for the ThreadPool that stores data at each step. This is
    needed because storing data takes a lot of time sometimes, even more than the computation itself.
    """
    def __init__(self, data_writer: DataWriter, n_threads: int):
        assert n_threads >= 0, n_threads
        self.data_writer = data_writer
        self.n_threads = min(n_threads, cpu_count())
        self.threads: list[Thread] = []
        if n_threads >= 1:
            self.queue: Queue = Queue(maxsize=1) # maxisze=1 because we use many shared memory threads.
            for _ in range(n_threads):
                self.threads.append(thr := Thread(target=self._worker_fn, daemon=True))
                thr.start()
        logger.debug(f"[{self.data_writer.rep.name}] Set up with {n_threads} threads.")

    def _worker_fn(self):
        while True: # This loop is ended via `self.join_with_timeout` or `__del__` or manually setting queue to None.
            if self.queue is None:
                logger.debug(f"[{self.data_writer.rep.name}] Queue of is None. Closing.")
                break
            try:
                args = self.queue.get(timeout=1)
            except (Empty, AttributeError): # AttributeError if it became 'None' during the if and the .get()
                continue
            self.data_writer(*args)
            self.queue.task_done()

    def join_with_timeout(self, timeout: int):
        """calls queue.join() but throws after timeout seconds if it doesn't end"""
        if self.n_threads == 0:
            return
        logger.debug(f"[{self.data_writer.rep.name}] Waiting for {self.queue.unfinished_tasks} "
                     "leftover enqueued tasks")
        assert self.queue is not None, "Queue was closed, create a new DataStorer object..."
        self.queue.all_tasks_done.acquire()
        try:
            endtime = time() + timeout
            while self.queue.unfinished_tasks:
                remaining = endtime - time()
                if remaining <= 0.0:
                    raise RuntimeError("Queue has not finished the join() before timeout")
                self.queue.all_tasks_done.wait(remaining)
        finally:
            self.queue.all_tasks_done.release()
        self.queue = None

    def __call__(self, y_repr: ReprOut):
        assert isinstance(y_repr, ReprOut), f"{self.data_writer.rep=}, {type(y_repr)=}"
        if self.n_threads == 0:
            self.data_writer(y_repr)
        else:
            assert self.queue is not None, "Queue was closed, create a new DataStorer object..."
            self.queue.put((deepcopy(y_repr), ), block=True, timeout=30)

    def __repr__(self):
        return f"""[DataStorer]
- Num threads: {self.n_threads} (0 = only using main thread)
{self.data_writer}"""

    def __del__(self):
        self.queue = None



"""Mask2Former representation"""
import json
from argparse import ArgumentParser, Namespace
from pathlib import Path
from datetime import datetime
from types import SimpleNamespace
from overrides import overrides
import torch as tr
import numpy as np

from vre.vre_video import VREVideo, FakeVideo
from vre.logger import vre_logger as logger
from vre.utils import image_resize_batch, image_read, image_write, MemoryData
from vre.representations import ReprOut, LearnedRepresentationMixin, ComputeRepresentationMixin
from vre_repository.weights_repository import fetch_weights
from vre_repository.semantic_segmentation import SemanticRepresentation

try:
    from .mask2former_impl import MaskFormer, CfgNode, get_output_shape
except ImportError:
    from vre_repository.semantic_segmentation.mask2former.mask2former_impl import MaskFormer, CfgNode, get_output_shape

class Mask2Former(SemanticRepresentation, LearnedRepresentationMixin, ComputeRepresentationMixin):
    """Mask2Former representation implementation. Note: only semantic segmentation (not panoptic/instance) enabled."""
    def __init__(self, model_id: str, disk_data_argmax: bool = False, **kwargs):
        LearnedRepresentationMixin.__init__(self)
        ComputeRepresentationMixin.__init__(self)
        assert isinstance(model_id, str) and model_id in {"47429163_0", "49189528_1", "49189528_0"}, model_id
        self._m2f_resources = Path(__file__).parent / "mask2former_impl/resources"
        classes, color_map, self.thing_dataset_id_to_contiguous_id = self._get_metadata(model_id)
        self.disk_data_argmax = disk_data_argmax
        SemanticRepresentation.__init__(self, classes=classes, color_map=color_map,
                                        disk_data_argmax=disk_data_argmax, **kwargs)
        self.model_id = model_id
        self.model: MaskFormer | None = None
        self.cfg: CfgNode | None = None
        self.output_dtype = "uint8" if disk_data_argmax else "float16"

    @property
    @overrides
    def n_channels(self) -> int:
        return len(self.classes)

    @tr.no_grad()
    @overrides
    def compute(self, video: VREVideo, ixs: list[int]):
        assert self.data is None, f"[{self}] data must not be computed before calling this"
        height, width = video.frame_shape[0:2]
        _os = get_output_shape(height, width, self.cfg.INPUT.MIN_SIZE_TEST, self.cfg.INPUT.MAX_SIZE_TEST)
        imgs = image_resize_batch(video[ixs], _os[0], _os[1], "bilinear", "PIL").transpose(0, 3, 1, 2).astype("float32")
        inputs = [{"image": tr.from_numpy(img), "height": height, "width": width} for img in imgs]
        predictions: list[tr.Tensor] = [x["sem_seg"] for x in self.model(inputs)]
        res = []
        for pred in predictions:
            res.append(pred.permute(1, 2, 0).to("cpu").numpy())
        self.data = ReprOut(frames=video[ixs], output=MemoryData(res), key=ixs)

    @staticmethod
    @overrides
    def weights_repository_links(**kwargs) -> list[str]:
        match kwargs["model_id"]:
            case "49189528_0": return ["semantic_segmentation/mask2former/49189528_0.ckpt"]
            case "49189528_1": return ["semantic_segmentation/mask2former/49189528_1.ckpt"]
            case "47429163_0": return ["semantic_segmentation/mask2former/47429163_0.ckpt"]
            case _: raise NotImplementedError(kwargs)

    @overrides
    def vre_setup(self, load_weights = True):
        assert self.setup_called is False
        self.cfg = CfgNode(json.load(open(f"{self._m2f_resources}/{self.model_id}_cfg.json", "r")))
        params = MaskFormer.from_config(self.cfg)
        params["metadata"] = SimpleNamespace(thing_dataset_id_to_contiguous_id=self.thing_dataset_id_to_contiguous_id)
        self.model = MaskFormer(**{**params, "semantic_on": True, "panoptic_on": False, "instance_on": False})
        if load_weights:
            ckpt_path = fetch_weights(Mask2Former.weights_repository_links(model_id=self.model_id))[0]
            ckpt_data = tr.load(ckpt_path, map_location="cpu")
            res = self.model.load_state_dict(ckpt_data["state_dict"], strict=False) # inference only: remove criterion
            assert res.unexpected_keys in (["criterion.empty_weight"], []), res
        self.model = self.model.eval().to(self.device)
        self.setup_called = True

    @overrides
    def vre_free(self):
        assert self.setup_called is True and self.model is not None, (self.setup_called, self.model is not None)
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None
        self.setup_called = False

    def _get_metadata(self, model_id: str) -> tuple[list[str], list[tuple[int, int, int]], dict[str, int]]:
        metadata = None
        if model_id == "49189528_1": # r50
            metadata = json.load(open(f"{self._m2f_resources}/mapillary_metadata.json", "r"))
        if model_id == "47429163_0": # swin
            metadata = json.load(open(f"{self._m2f_resources}/coco_metadata.json", "r"))
        if model_id == "49189528_0": # swin
            metadata = json.load(open(f"{self._m2f_resources}/mapillary_metadata2.json", "r"))
        return metadata["stuff_classes"], metadata["stuff_colors"], metadata.get("thing_dataset_id_to_contiguous_id")

def get_args() -> Namespace:
    """cli args"""
    parser = ArgumentParser()
    parser.add_argument("model_id", choices=["49189528_1", "47429163_0", "49189528_0"])
    parser.add_argument("input_image", type=Path)
    parser.add_argument("output_path", type=Path)
    return parser.parse_args()

def main(args: Namespace):
    """main fn. Usage: python mask2former.py 49189528_1/47429163_0/49189528_0 demo1.jpg output1.jpg"""
    img = image_read(args.input_image)

    m2f = Mask2Former(args.model_id, disk_data_argmax=False, name="m2f", dependencies=[])
    m2f.device = "cuda" if tr.cuda.is_available() else "cpu"
    m2f.vre_setup()
    now = datetime.now()
    m2f.compute(FakeVideo(img[None], 1), [0])
    logger.info(f"Pred took: {datetime.now() - now}")
    semantic_result: np.ndarray = m2f.make_images(m2f.data)[0]
    image_write(semantic_result, args.output_path)
    logger.info(f"Written prediction to '{args.output_path}'")
    return m2f.data.output # for integration tests

if __name__ == "__main__":
    main(get_args())



# pylint: disable=all
import torch
import torch.nn.functional as F


def median_filter(x, kernel_size=3):
    nb, nc, ny, nx = x.shape
    x = x.reshape(nb * nc, 1, ny, nx)
    x = F.unfold(x, kernel_size, padding=(kernel_size - 1) // 2)
    y = x.median(dim=1, keepdim=True)[0]
    y = F.fold(y, (ny, nx), 1)
    y = y.view(nb, nc, ny, nx)
    return y


def like_an_image(seg, img, sampling_step=5):
    nb, nc, ny, nx = seg.shape
    ni = img.shape[1]
    img = F.interpolate(img, size=(ny, nx), mode='nearest')
    seg = torch.cat((seg, torch.ones((nb, 1, seg.shape[2], seg.shape[3]), dtype=seg.dtype, device=seg.device)), dim=1)

    ys, xs = torch.meshgrid(torch.arange(0, ny - sampling_step, sampling_step),
                            torch.arange(0, nx - sampling_step, sampling_step), indexing="ij")
    seg_spl = seg[:, :, ys, xs].view(nb, nc + 1, -1).permute(0, 2, 1).unsqueeze(1).expand(-1, ni, -1, -1)
    img_spl = img[:, :, ys, xs].view(nb, ni, -1).unsqueeze(3)

    trf = torch.linalg.lstsq(seg_spl, img_spl)[0]
    seg_img = seg.view(nb, nc + 1, -1).permute(0, 2, 1).unsqueeze(1).expand(-1, ni, -1, -1)
    seg_img = seg_img @ trf
    seg_img = seg_img.squeeze(3).view(nb, ni, ny, nx)

    return seg_img


def rgb2hsv(img, epsilon=1e-10):
    r, g, b = img[:, 0], img[:, 1], img[:, 2]
    max_rgb, _ = img.max(1)
    min_rgb, argmin_rgb = img.min(1)

    max_min = max_rgb - min_rgb + epsilon

    h1 = 60.0 * (g - r) / max_min + 60.0
    h2 = 60.0 * (b - g) / max_min + 180.0
    h3 = 60.0 * (r - b) / max_min + 300.0

    h = torch.stack((h2, h3, h1), dim=0).gather(dim=0, index=argmin_rgb.unsqueeze(0)).squeeze(0)
    s = max_min / (max_rgb + epsilon)
    v = max_rgb

    return torch.stack((h / 360, s, v), dim=1)


def hsv2rgb(img):
    assert(img.shape[1] == 3)

    h, s, v = img[:, 0] * 360, img[:, 1], img[:, 2]
    h_ = (h - torch.floor(h / 360) * 360) / 60
    c = s * v
    x = c * (1 - torch.abs(torch.fmod(h_, 2) - 1))

    zero = torch.zeros_like(c)
    y = torch.stack((
        torch.stack((c, x, zero), dim=1),
        torch.stack((x, c, zero), dim=1),
        torch.stack((zero, c, x), dim=1),
        torch.stack((zero, x, c), dim=1),
        torch.stack((x, zero, c), dim=1),
        torch.stack((c, zero, x), dim=1),
    ), dim=0)

    index = torch.repeat_interleave(torch.floor(h_).unsqueeze(1), 3, dim=1).unsqueeze(0).to(torch.long)
    rgb = (y.gather(dim=0, index=index) + (v - c)).squeeze(0)
    return rgb


def normalize_channels(x, whole_block=False):
    _, nc, ny, nx = x.shape
    if whole_block:
        min_c = x.view(x.shape[0], nc * ny * nx).min(dim=1, keepdim=True)[0].unsqueeze(-1).unsqueeze(-1)
        max_c = x.view(x.shape[0], nc * ny * nx).max(dim=1, keepdim=True)[0].unsqueeze(-1).unsqueeze(-1)
    else:
        min_c = x.view(x.shape[0], -1, ny * nx).min(dim=2, keepdim=True)[0].unsqueeze(-1)
        max_c = x.view(x.shape[0], -1, ny * nx).max(dim=2, keepdim=True)[0].unsqueeze(-1)
    return (x - min_c) / (max_c - min_c + torch.finfo(x.dtype).eps)




"""FFmpegVideo module"""
from pathlib import Path
import ffmpeg
import numpy as np
from overrides import overrides
from tqdm import trange

from .vre_video import VREVideo

class FFmpegVideo(VREVideo):
    """FFmpegVideo -- reads data from a video using ffmpeg"""
    def __init__(self, path: Path, cache_len: int = 30):
        super().__init__()
        self.path = Path(path)
        assert self.path.exists(), f"Video '{self.path}' doesn't exist"
        self.probe = ffmpeg.probe(self.path)
        self.stream_info = next((stream for stream in self.probe["streams"] if stream["codec_type"] == "video"), None)
        self.width = int(self.stream_info["width"])
        self.height = int(self.stream_info["height"])
        self.frame_shape = (self.height, self.width, 3)
        self._fps = eval(self.stream_info["avg_frame_rate"]) # pylint: disable=eval-used
        self.total_frames = self._build_total_frames()

        self.cache = []
        self.cache_max_len = cache_len
        self.cache_start_frame = None
        self.cache_end_frame = None
        self.video_process = None

    @property
    @overrides
    def shape(self) -> tuple[int, int, int, int]:
        return (self.total_frames, *self.frame_shape)

    @property
    @overrides
    def fps(self) -> float:
        return self._fps

    @overrides
    def write(self, out_path: Path, start_frame: int = 0, end_frame: int | None = None):
        out_path = Path(out_path)
        assert self.write_process is None, self.write_process
        assert out_path.suffix == ".mp4", out_path
        assert isinstance(start_frame, int) and start_frame >= 0, start_frame

        self.write_process = (
            ffmpeg
            .input("pipe:0", format="rawvideo", pix_fmt="rgb24", s=f"{self.shape[2]}x{self.shape[1]}", r=self.fps)
            .output(str(out_path), pix_fmt="yuv420p", vcodec="libx264")
            .overwrite_output()
            .run_async(pipe_stdin=True, pipe_stderr=-3, pipe_stdout=-3) # -3 = subprocess.DEVNULL
        )

        try:
            for frame_ix in trange(start_frame, end_frame or len(self)):
                self.write_process.stdin.write(self[frame_ix].tobytes())
        finally:
            self.write_process.stdin.close()
            self.write_process.wait()
            self.write_process = None

    def _build_total_frames(self) -> int:
        """returns the number of frames of the vifdeo"""
        if "nb_frames" in self.stream_info:
            return int(float(self.stream_info["nb_frames"]))
        if "codec_name" in self.stream_info and self.stream_info["codec_name"] == "h264":
            if "tags" in self.stream_info and "DURATION" in self.stream_info["tags"]:
                duration_str = self.stream_info["tags"]["DURATION"]
                h, m, s = [float(x) for x in duration_str.split(":")]
                duration_s = h * 60 * 60 + m * 60 + s
                return int(duration_s * self.fps)
        raise ValueError(f"Unknown video format. Stream info from ffmpeg: {self.stream_info}")

    def _start_ffmpeg_process(self, start_time):
        """
        Start an ffmpeg process from the nearest keyframe to the requested start_time.
        This will load all frames from this keyframe to the next keyframe.
        """
        # Close any existing process
        if self.video_process is not None:
            self.video_process.stdout.close()
            self.video_process.stderr.close()
            self.video_process.terminate()

        # Seek to the nearest keyframe before `start_time`
        self.video_process = (
            ffmpeg
            .input(self.path, ss=start_time)
            .output("pipe:", format="rawvideo", pix_fmt="rgb24")
            .run_async(pipe_stdout=True, pipe_stderr=True, pipe_stdin=True)
        )

    def _cache_frames(self, start_frame: int):
        """Cache all frames between the current keyframe and the next keyframe, starting at start_frame."""
        start_time = start_frame / self.fps
        self.cache = []
        self._start_ffmpeg_process(start_time)

        # Read frames until the end of the current keyframe range
        while True:
            in_bytes = self.video_process.stdout.read(self.width * self.height * 3)
            if not in_bytes or len(self.cache) > self.cache_max_len:
                break
            frame = np.frombuffer(in_bytes, np.uint8).reshape([self.height, self.width, 3])
            self.cache.append(frame)

        # Set cache boundaries in terms of frame numbers
        self.cache_start_frame = start_frame
        self.cache_end_frame = start_frame + len(self.cache)

    def get_frame_by_number(self, frame_number: int) -> np.ndarray:
        """Retrieve a frame from the video by frame number, using nearby frames caching."""
        assert isinstance(frame_number, int), type(frame_number)
        assert 0 <= frame_number < self.total_frames, f"Frame out of bounds: {frame_number}. Len: {len(self)}"

        # Load new cache if the requested frame is outside the current cache range
        if self.cache_start_frame is None or not self.cache_start_frame <= frame_number < self.cache_end_frame:
            self._cache_frames(frame_number)

        # Calculate the index within the cache
        return self.cache[frame_number - self.cache_start_frame]

    def __repr__(self):
        return f"[FFmpegVideo] Path: {self.path}. FPS: {self.fps}. Len: {len(self)}. Frame shape: {self.frame_shape}."

    def __len__(self) -> int:
        return self.total_frames

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        if isinstance(ix, np.ndarray):
            return self[ix.tolist()]
        if isinstance(ix, list):
            return np.array([self[_ix] for _ix in ix])
        if isinstance(ix, slice):
            return np.array([self[_ix] for _ix in range(ix.start, ix.stop)])
        return self.get_frame_by_number(ix)

    def __del__(self):
        """Clean up the ffmpeg process when done."""
        if hasattr(self, "video_process") and self.video_process is not None: # in case it throws in the constructor
            self.video_process.stdout.close()
            self.video_process.stderr.close()
            self.video_process.terminate()



"""optical_flow_representationpy -- module implementing an Optical Flow Represenatation generic class"""
from overrides import overrides
import numpy as np
from vre.utils import colorize_optical_flow, ReprOut
from vre.representations import Representation, NpIORepresentation, NormedRepresentationMixin

class OpticalFlowRepresentation(Representation, NpIORepresentation, NormedRepresentationMixin):
    """OpticalFlowRepresentation. Implements flow task-specific stuff."""
    def __init__(self, name: str, **kwargs):
        Representation.__init__(self, name, **kwargs)
        NpIORepresentation.__init__(self)
        NormedRepresentationMixin.__init__(self)

    @property
    @overrides
    def n_channels(self) -> int:
        return 2

    @overrides
    def make_images(self, data: ReprOut) -> np.ndarray:
        assert self.data is not None, f"[{self}] data must be first computed using compute()"
        y = self.unnormalize(data.output) if self.normalization is not None else self.data.output
        return colorize_optical_flow(y)



"""external_representations.py handles functions related to loading external representations to the library"""
import imp # pylint: disable=deprecated-module
from pathlib import Path
from typing import Type
from omegaconf import DictConfig, OmegaConf
from .representation import Representation
from .io_representation_mixin import IORepresentationMixin
from .compute_representation_mixin import ComputeRepresentationMixin
from .learned_representation_mixin import LearnedRepresentationMixin
from ..logger import vre_logger as logger
from ..utils import topological_sort

def add_external_repositories(external_paths: list[str],
                              default_representations: dict[str, type[Representation]] | None = None) \
        -> dict[str, type[Representation]]:
    """adds external repositories from provided paths in the format: /path/to/script.py:fn_name"""
    representation_types = default_representations or {}
    for external_path in external_paths:
        external_types = _add_one_external_repository(external_path)
        assert (diff := set(representation_types.keys()).intersection(external_types)) == set(), diff
        representation_types = {**representation_types, **external_types}
    return representation_types

def _add_one_external_repository(external_path: str) -> dict[str, type[Representation]]:
    path, fn = external_path.split(":")
    external_reprs: dict[str, Representation] = getattr(imp.load_source("external", path), fn)()
    assert all(isinstance(v, Type) for v in external_reprs.values()), external_reprs
    return external_reprs

def _validate_repr_cfg(repr_cfg: dict, name: str):
    assert isinstance(repr_cfg, dict), f"Broken format (not a dict) for {name}. Type: {type(repr_cfg)}."
    valid_keys = {"type", "parameters", "dependencies", "compute_parameters", "learned_parameters", "io_parameters"}
    required_keys = {"type", "parameters", "dependencies"}
    assert set(repr_cfg).issubset(valid_keys), f"{name} wrong keys: {repr_cfg.keys()}"
    assert (diff := required_keys.difference(repr_cfg)) == set(), f"{name} missing keys: {diff}"
    assert isinstance(repr_cfg["parameters"], dict), type(repr_cfg["parameters"])
    assert name.find("/") == -1, "No '/' allowed in the representation name. Got '{name}'"

def build_representation_from_cfg(repr_cfg: dict, name: str, representation_types: dict[str, type[Representation]],
                                  built_so_far: list[Representation],
                                  compute_defaults: dict, learned_defaults: dict, io_defaults: dict) -> Representation:
    """Builds a representation given a dict config and a name."""
    assert isinstance(learned_defaults, dict) and isinstance(compute_defaults, dict) and isinstance(io_defaults, dict)
    _validate_repr_cfg(repr_cfg, name)

    logger.info(f"Building '{repr_cfg['type']}' (vre name: {name})")
    obj_type = representation_types[repr_cfg["type"]]
    built_so_far_dict = {r.name: r for r in built_so_far}
    dependencies = [built_so_far_dict[dep] for dep in repr_cfg["dependencies"]]
    obj: Representation = obj_type(name=name, dependencies=dependencies, **repr_cfg["parameters"])

    learned_params, compute_params, io_params = learned_defaults, compute_defaults, io_defaults
    if "learned_parameters" in repr_cfg:
        assert isinstance(obj, LearnedRepresentationMixin), obj
        learned_params = {**learned_params, **repr_cfg["learned_parameters"]}
        logger.debug(f"[{obj}] Setting node specific 'Learned' params: {learned_params}")
    if "compute_parameters" in repr_cfg:
        assert isinstance(obj, ComputeRepresentationMixin), obj
        compute_params = {**compute_params, **repr_cfg["compute_parameters"]}
        logger.debug(f"[{obj}] Setting node specific 'Compute' params: {compute_params}")
    if "io_parameters" in repr_cfg:
        assert isinstance(obj, IORepresentationMixin), obj
        io_params = {**io_params, **repr_cfg["io_parameters"]}
        logger.debug(f"[{obj}] Setting node specific 'IO' params: {io_params}")

    if isinstance(obj, ComputeRepresentationMixin):
        obj.set_compute_params(**compute_params)
    if isinstance(obj, LearnedRepresentationMixin):
        obj.set_learned_params(**learned_params)
    if isinstance(obj, IORepresentationMixin):
        obj.set_io_params(**io_params)
    return obj

def build_representations_from_cfg(cfg: Path | str | DictConfig | dict,
                                   representation_types: dict[str, type[Representation]],
                                   external_representations: list[Path] | None = None) -> list[Representation]:
    """builds a list of representations given a dict config (yaml file)"""
    assert isinstance(cfg, (Path, str, DictConfig, dict)), type(cfg)
    cfg = OmegaConf.load(cfg) if isinstance(cfg, (Path, str)) else cfg
    cfg: dict = OmegaConf.to_container(cfg, resolve=True) if isinstance(cfg, DictConfig) else cfg
    assert len(repr_cfg := cfg["representations"]) > 0 and isinstance(repr_cfg, dict), repr_cfg

    logger.debug("Doing topological sort...")
    dep_graph = {repr_name: repr_cfg_values["dependencies"] for repr_name, repr_cfg_values in repr_cfg.items()}
    topo_sorted = {k: repr_cfg[k] for k in topological_sort(dep_graph)}

    compute_defaults = cfg.get("default_compute_parameters", {})
    learned_defaults = cfg.get("default_learned_parameters", {})
    io_defaults = cfg.get("default_io_parameters", {})

    built_so_far: list[Representation] = []
    for name, repr_cfg in topo_sorted.items():
        obj = build_representation_from_cfg(repr_cfg=repr_cfg, name=name, representation_types=representation_types,
                                            built_so_far=built_so_far, compute_defaults=compute_defaults,
                                            learned_defaults=learned_defaults, io_defaults=io_defaults)
        built_so_far.append(obj)

    for external_repr in (external_representations or []):
        built_so_far = _add_one_external_representation_list(built_so_far, external_repr,
                                                             compute_params=compute_defaults,
                                                             learned_params=learned_defaults, io_params=io_defaults)
    return built_so_far

def _add_external_representations_dict(built_so_far: list[Representation],
                                       external_reprs: dict[str, Representation],
                                       compute_params: dict, learned_params: dict,
                                       io_params: dict) -> list[Representation]:
    assert all(isinstance(v, Representation) for v in external_reprs.values()), external_reprs
    name_to_repr = {r.name: r for r in built_so_far}
    assert (diff := set(name_to_repr.keys()).intersection(external_reprs)) == set(), diff

    # Note (TODO?): ideally we'd use deepcopy here so we don't update external_reprs stuff, however there is an issue
    # if we do this, becvause there is no guarantee that external_reprs is topo-sorted (which we could do). For this
    # reason. we cannot do stuff like semantic_output (see semantic_mapper) that depends on 3 representations
    # (converted) which are also task mapped from the raw ones, because we'd update the dependencies of the middle ones
    # (converted) initially os it uses 'built_so_far' objects, but then it crashes on depth=2 (semantic_output).
    for obj in external_reprs.values():
        if isinstance(obj, ComputeRepresentationMixin):
            obj.set_compute_params(**compute_params)
        if isinstance(obj, LearnedRepresentationMixin):
            obj.set_learned_params(**learned_params)
        if isinstance(obj, IORepresentationMixin):
            obj.set_io_params(**io_params)

        # update clashes in dependencies
        for i, external_dep in enumerate(obj.dependencies):
            if external_dep.name in name_to_repr and id(external_dep) != id(curr := name_to_repr[external_dep.name]):
                logger.warning(f"[{obj.name}] Dependency {external_dep} is different than existing {curr}. "
                                "Replacing the dependency. This may yield in wrong results!")
                obj.dependencies[i] = curr
    return [*built_so_far, *external_reprs.values()]

def _add_one_external_representation_list(built_so_far: list[Representation], external_path: str, compute_params: dict,
                                          learned_params: dict, io_params: dict) -> list[Representation]:
    assert isinstance(learned_params, dict) and isinstance(compute_params, dict) and isinstance(io_params, dict)
    path, fn = external_path.split(":")
    external_reprs: dict[str, Representation] = getattr(imp.load_source("external", path), fn)()
    assert isinstance(external_reprs, dict) and len(external_reprs) > 0, (external_path, external_reprs)
    logger.info(f"Adding {list(external_reprs)} from {path}")
    return _add_external_representations_dict(built_so_far, external_reprs, compute_params, learned_params, io_params)



#!/usr/bin/env python3
"""MultiTask Dataset module compatible with torch.utils.data.Dataset & DataLoader."""
from __future__ import annotations
import os
from pathlib import Path
from typing import Iterable, NamedTuple
from pprint import pformat
from copy import deepcopy
from natsort import natsorted
import torch as tr
from torch.utils.data import Dataset

from vre.representations import Representation, NormedRepresentationMixin, IORepresentationMixin, TaskMapper
from vre.logger import vre_logger as logger

from .statistics import compute_statistics, load_external_statistics, TaskStatistics

BuildDatasetTuple = NamedTuple("build_dataset_res", [("files_per_task", dict[list[Path]]), ("all_names", list[str]),
                                                     ("task_types", dict[str, type[Representation]])])
"""
A tuple of two items:
- files_per_repr A dict {task: [list of disk data]} where files_per_repr["rgb"][0] -> /path/to/0.npz
    - Additionally, if a representation is a TaskMapper, it'll return [/path/to/dep1/0.npz, /path/to/depn/0.npz]
- all_names A list without prefix paths i.e. [0.npz, 1.npz, ..., n.npz]. It supports non numeric frame numbers as well.
- task_types A dictionary for each type of a representation with the key being its task name.
"""

MultiTaskItem = tuple[dict[str, tr.Tensor], str] # [{task: data}, stem(name) | list[stem(name)]]
Repr = Representation | IORepresentationMixin

class MultiTaskDataset(Dataset):
    """
    MultiTaskDataset implementation. Reads data from npz files and returns them as a dict.
    Parameters:
    - path: Path to the directory containing the npz files.
    - task_names: List of tasks that are present in the dataset. If set to None, will infer from the files on disk.
    - task_types: A dictionary of form {task_name: task_type} for the reader to call to read from disk, plot etc.
    - normalization: The normalization type used in __getitem__. Valid options are:
      - None: Reads the data as-is using task.read_from_disk(path)
      - 'min_max': Calls task.normalize(task.read_from_disk(path), mins[task], maxs[task])
      - 'standardization': Calls task.standardize(task.read_from_disk(path), means[task], stds[task])
      If normalization is not 'none', then task-level statistics will be computed. Environmental variable
      STATS_PBAR=0/1 enables tqdm during statistics computation.
    - handle_missing_data: Modes to handle missing data. Defaults to 'fill_none'. Valid options are:
      - 'raise': Raise exception if any missing data.
      - 'drop': Drop the data point if any of the representations is missing.
      - 'fill_{none,zero,nan}': Fill the missing data with Nones, zeros or NaNs.
    - files_suffix: What suffix to look for when creating the dataset. Valid values: 'npy' or 'npz'.
    - cache_task_stats: If set to True, the statistics will be cached at '{path}/.task_statistics.npz'. Can be enabled
    using the environmental variable STATS_CACHE=1. Defaults to False.
    - batch_size_stats: Controls the batch size during statistics computation. Can be enabled by environmental variable
    STATS_BATCH_SIZE. Defaults to 1.
    - num_workers_stats: Controls the num_woprkers during statistics computation. Can be enabled by environmental
    variable STATS_NUM_WORKERS. Defaults to 0.
    - statistics The dictionary of statistics which can be externally provided too, otherwise computes or loads them
    from the datasert dir if normalization is set.

    Expected directory structure:
    path/
    - task_1/0.npz, ..., N.npz
    - ...
    - task_n/0.npz, ..., N.npz

    Names can be in a different format (i.e. 2022-01-01.npz), but must be consistent and equal across all tasks.
    """

    def __init__(self, path: Path,
                 task_names: list[str],
                 task_types: dict[str, type],
                 normalization: str | None | dict[str],
                 handle_missing_data: str = "fill_none",
                 files_suffix: str = "npz",
                 cache_task_stats: bool = (os.getenv("STATS_CACHE", "1") == "1"),
                 batch_size_stats: int = int(os.getenv("STATS_BATCH_SIZE", "1")),
                 num_workers_stats: int = int(os.getenv("NUM_WORKERS_STATS", "0")),
                 statistics: dict[str, TaskStatistics] | None = None,
    ):
        assert Path(path).exists(), f"Provided path '{path}' doesn't exist!"
        assert handle_missing_data in ("drop", "fill_none", "fill_zero", "fill_nan", "raise"), \
            f"Invalid handle_missing_data mode: {handle_missing_data}"
        assert isinstance(task_names, Iterable) and isinstance(task_types, dict), (type(task_names), type(task_types))
        self.path = Path(path).absolute()
        self.handle_missing_data = handle_missing_data
        self.suffix = files_suffix
        self.files_per_repr, self.file_names, self.task_types = self._build_dataset(task_types, task_names)
        self.task_names = sorted(task_names)
        self.cache_task_stats = cache_task_stats
        self.batch_size_stats = batch_size_stats
        self.num_workers_stats = num_workers_stats

        logger.info(f"Tasks used in this dataset: {self.task_names}")

        self._data_shape: tuple[int, ...] | None = None
        self._tasks: list[Repr] | None = None
        self._default_vals: dict[str, tr.Tensor] | None = None
        self._statistics: dict[str, TaskStatistics] | None = None
        if statistics is not None:
            self._statistics = load_external_statistics(self, statistics)

        if normalization is not None:
            if isinstance(normalization, str):
                logger.debug(f"Normalization provided as a string ({normalization}). Setting all tasks to this")
                _norm, normalization = normalization, {}
                for tn in self.task_names:
                    normalization[tn] = None if self.name_to_task[tn] in self.classification_tasks else _norm
            if "*" in normalization.keys(): # for the lazy, we can put {"*": "standardization", "depth": "min_max"}
                value = normalization.pop("*")
                for missing_task in set(self.task_names).difference(normalization.keys()):
                    normalization[missing_task] = value
            assert all(n in ("min_max", "standardization") or n is None for n in normalization.values()), normalization
            assert all(k in task_names for k in normalization.keys()), set(normalization).difference(task_names)
        self.normalization: dict[str, str] | None = normalization

    # Public methods and properties

    @property
    def statistics(self) -> dict[str, TaskStatistics] | None:
        """returns the statistics of this dataset for all NormedRepresentation tasks, None otherwise"""
        if self.normalization is None:
            return None
        if self._statistics is None:
            self._statistics = compute_statistics(self)
        for task_name, task in self.name_to_task.items():
            if not hasattr(task, "classes") and task.normalization is None: # TODO: used NormedRepresentationMixin
                task.set_normalization(self.normalization[task_name], self._statistics[task_name])
        return self._statistics

    @property
    def name_to_task(self) -> dict[str, Repr]:
        """A dict that maps the name of the task to the task"""
        return {task.name: task for task in self.tasks}

    @property
    def default_vals(self) -> dict[str, tr.Tensor]:
        """default values for __getitem__ if item is not on disk but we retrieve a full batch anyway"""
        _default_val = float("nan") if self.handle_missing_data == "fill_nan" else 0
        return {task: None if self.handle_missing_data == "fill_none" else tr.full(self.data_shape[task], _default_val)
                for task in self.task_names}

    @property
    def data_shape(self) -> dict[str, tuple[int, ...]]:
        """Returns a {task: shape_tuple} for all representations. At least one npz file must exist for each."""
        first_npz = {task: [_v for _v in files if _v is not None][0] for task, files in self.files_per_repr.items()}
        data_shape = {}
        for task_name, task in self.name_to_task.items():
            if isinstance(task, TaskMapper) and task.dependencies[0] != task:
                data_shape[task_name] = task.compute_from_dependencies_paths(first_npz[task_name]).shape
            else:
                data_shape[task_name] = task.disk_to_memory_fmt(task.load_from_disk(first_npz[task_name])).shape
        return data_shape

    @property
    def mins(self) -> dict[str, tr.Tensor]:
        """returns a dict {task: mins[task]} for all the tasks if self.statistics exists"""
        assert self.normalization is not None, "No statistics for normalization is None"
        return {k: v[0] for k, v in self.statistics.items() if k in self.task_names}

    @property
    def maxs(self) -> dict[str, tr.Tensor]:
        """returns a dict {task: mins[task]} for all the tasks if self.statistics exists"""
        assert self.normalization is not None, "No statistics for normalization is None"
        return {k: v[1] for k, v in self.statistics.items() if k in self.task_names}

    @property
    def means(self) -> dict[str, tr.Tensor]:
        """returns a dict {task: mins[task]} for all the tasks if self.statistics exists"""
        assert self.normalization is not None, "No statistics for normalization is None"
        return {k: v[2] for k, v in self.statistics.items() if k in self.task_names}

    @property
    def stds(self) -> dict[str, tr.Tensor]:
        """returns a dict {task: mins[task]} for all the tasks if self.statistics exists"""
        assert self.normalization is not None, "No statistics for normalization is None"
        return {k: v[3] for k, v in self.statistics.items() if k in self.task_names}

    @property
    def tasks(self) -> list[Repr]:
        """
        Returns a list of instantiated tasks in the same order as self.task_names. Overwrite this to add
        new tasks and semantics (i.e. plot_fn or doing some preprocessing after loading from disk in some tasks.
        """
        if self._tasks is None:
            self._tasks = []
            for task_name in self.task_names:
                self._tasks.append(self.task_types[task_name])
            assert all(t.name == t_n for t, t_n in zip(self._tasks, self.task_names)), (self.task_names, self._tasks)
        return self._tasks

    @property
    def classification_tasks(self) -> list[Repr]:
        """Returns a list of only the classification tasks"""
        return [task for task in self.tasks if task.is_classification]

    def collate_fn(self, items: list[MultiTaskItem]) -> MultiTaskItem:
        """
        given a list of items (i.e. from a reader[n:n+k] call), return the item batched on 1st dimension.
        Nones (missing data points) are turned into nans as per the data shape of that dim.
        """
        items_name = [item[1] for item in items]
        res = {k: tr.zeros(len(items), *self.data_shape[k]).float() for k in self.task_names} # float32 always
        for i in range(len(items)):
            for k in self.task_names:
                try:
                    res[k][i][:] = items[i][0][k] if items[i][0][k] is not None else float("nan")
                except Exception as e:
                    print(k, items)
                    raise e
        return res, items_name

    def add_task(self, task: Repr, overwrite: bool=False):
        """Safely adds a task to this reader. Most likely can be optimized"""
        logger.info(f"Adding a new task: '{task.name}'")
        if task.name in self.task_names:
            if overwrite:
                self.remove_task(task.name)
            else:
                raise ValueError(f"Task '{task.name}' already exists: {self.task_names}")
        self.task_names = sorted([*self.task_names, task.name])
        self.task_types[task.name] = task
        self._tasks = None
        self.files_per_repr, self.file_names, self.task_types = self._build_dataset(self.task_types, self.task_names)

    def remove_task(self, task_name: str):
        """Safely removes a task from this reader"""
        logger.info(f"Removing a task: '{task_name}'")
        assert task_name in self.task_names, f"Task '{task_name}' doesn't exist: {self.task_names}"
        self.task_names = sorted(name for name in self.task_names if name != task_name)
        del self.task_types[task_name]
        self._tasks = None
        self.files_per_repr, self.file_names, self.task_types = self._build_dataset(self.task_types, self.task_names)

    def get_one_item(self, index: int, subset_tasks: list[str] | None = None) -> MultiTaskItem:
        """getitem implementation for a single index. Supports a subset of tasks as well"""
        assert isinstance(index, int), type(index)
        assert isinstance(subset_tasks, (list, type(None))), type(subset_tasks)
        subset_tasks = self.task_names if subset_tasks is None else subset_tasks
        assert len(subset_tasks) > 0, subset_tasks
        res: dict[str, tr.Tensor] = {}
        for task_name in subset_tasks:
            assert task_name in self.name_to_task, (task_name, list(self.name_to_task))
            task = self.name_to_task[task_name]
            file_path = self.files_per_repr[task_name][index]
            if file_path is None:
                res[task_name] = self.default_vals[task_name]
            else:
                if isinstance(task, TaskMapper) and task.dependencies[0] != task:
                    np_memory_data = task.compute_from_dependencies_paths(file_path)
                else: # can also be TaskMapper here too, but with deps[0] == task (pre-computed)
                    np_memory_data = task.disk_to_memory_fmt(task.load_from_disk(file_path))

                if self.statistics is not None and not hasattr(task, "classes"): # TODO: use NormedRepresentationMixin
                    assert isinstance(task, NormedRepresentationMixin), task
                    np_memory_data = task.normalize(np_memory_data)
                res[task_name] = tr.from_numpy(np_memory_data)
        return res, self.file_names[index]

    # Private methods

    def _get_all_npz_files(self) -> dict[str, list[Path]]:
        """returns a dict of form: {"rgb": ["0.npz", "1.npz", ..., "N.npz"]}"""
        assert self.suffix == "npz", f"Only npz supported right now (though trivial to update): {self.suffix}"
        in_files = {}
        all_repr_dirs: list[str] = [x.name for x in self.path.iterdir() if x.is_dir() and not x.name.startswith(".")]
        for repr_dir_name in all_repr_dirs:
            dir_name = self.path / repr_dir_name
            if all(f.is_dir() for f in dir_name.iterdir()): # dataset is stored as repr/part_x/0.npz, ..., part_k/n.npz
                all_files = []
                for part in dir_name.iterdir():
                    all_files.extend(part.glob(f"*.{self.suffix}"))
            else: # dataset is stored as repr/0.npz, ..., repr/n.npz
                all_files = dir_name.glob(f"*.{self.suffix}")
            all_files = [x for x in all_files if not x.name.endswith("_extra.npz")] # important: remove xxx_extra.npz
            in_files[repr_dir_name] = natsorted(all_files, key=lambda x: x.name) # important: use natsorted() here
        assert not any(len(x) == 0 for x in in_files.values()), f"{ [k for k, v in in_files.items() if len(v) == 0] }"
        return in_files

    def _build_dataset(self, task_types: dict[str, Repr], task_names: list[str]) -> BuildDatasetTuple:
        """Builds the dataset from the disk. Also handles missing data (i.e. looking at TaskMappings)."""
        logger.debug(f"Building dataset from: '{self.path}'")
        all_npz_files = self._get_all_npz_files()
        all_files: dict[str, dict[str, Path]] = {k: {_v.name: _v for _v in v} for k, v in all_npz_files.items()}
        task_types = deepcopy(task_types)

        assert (diff := set(task_names).difference(task_types)) == set(), f"\n-{diff}\n-{list(task_types)}"
        if (diff := set(task_names).difference(all_files)) != set():
            logger.debug(f"The following tasks do not have data on disk: {list(diff)}. Checking dependencies.")
        relevant_tasks_for_files = set() # fo e.g. hsv requires only rgb, so we look at dependencies later on
        for task_name in task_names:
            if task_name not in diff and task_types[task_name].dep_names != [task_name]:
                logger.debug(f"Upating the deps of '{task_name}' as all its data is on disk!")
                task_types[task_name].dependencies = [task_types[task_name]]
            relevant_tasks_for_files.update(task_types[task_name].dep_names)
        if (diff := relevant_tasks_for_files.difference(all_files)) != set():
            raise FileNotFoundError(f"Missing files for {diff}.\nFound on disk: {[*all_files]}\n{pformat(task_types)}")

        names_to_tasks: dict[str, list[str]] = {} # {name: [task]}
        for task_name in relevant_tasks_for_files: # just the relevant tasks
            for path_name in all_files[task_name].keys():
                names_to_tasks.setdefault(path_name, [])
                names_to_tasks[path_name].append(task_name)

        if self.handle_missing_data == "raise":
            diffs = []
            for k, _v in names_to_tasks.items():
                if (v := set(_v)) != (first := set(names_to_tasks[next(iter(names_to_tasks))])):
                    diffs.append((k, v, first, v - first, first - v))
            if len(diffs) > 0:
                for k, v, first, diff1, diff2 in diffs:
                    logger.error(f"Key '{k}' has different files:\n-{v=}\n-{first=}\n-{diff1=}\n-{diff2=}")
                raise ValueError

        if self.handle_missing_data == "drop":
            b4 = len(names_to_tasks)
            names_to_tasks = {k: v for k, v in names_to_tasks.items() if len(v) == len(relevant_tasks_for_files)}
            logger.debug(f"Dropped {b4 - len(names_to_tasks)} files not in all tasks")
        all_names: list[str] = natsorted(names_to_tasks.keys())
        logger.debug(f"Total files: {len(names_to_tasks)} per task across {len(task_names)} tasks")

        files_per_task: dict[str, list[str | None] | list[list[str] | None]] = {task: [] for task in task_names}
        for name in all_names:
            for task in task_names:
                all_deps_exist = set(deps := task_types[task].dep_names).issubset(names_to_tasks[name])
                if not all_deps_exist:
                    files_per_task[task].append(None) # if any of the deps don't exist for this task, skip it.
                else:
                    paths = [all_files[dep][name] for dep in deps]
                    assert len(paths) > 0, f"'{task}' most likely has no data or dependencies to compute it from"
                    files_per_task[task].append(paths if len(deps) > 1 else paths[0])

        assert all(isinstance(x, str) for x in task_names), tuple(zip(task_names, (type(x) for x in task_names)))
        assert (diff := set(files_per_task).difference(task_names)) == set(), f"Not all tasks in files: {diff}"
        return BuildDatasetTuple(files_per_task=files_per_task, all_names=all_names,
                                 task_types={k: v for k, v in task_types.items() if k in task_names})

    # Python magic methods (pretty printing the reader object, reader[0], len(reader) etc.)
    def __getitem__(self, index: int | str | slice | list[int, str] | tuple[int, str]) -> MultiTaskItem:
        """Read the data all the desired nodes"""
        assert isinstance(index, (int, slice, list, tuple, str)), type(index)
        if isinstance(index, slice):
            assert index.start is not None and index.stop is not None and index.step is None, "Only reader[l:r] allowed"
            index = list(range(index.start, index.stop))
        if isinstance(index, (list, tuple)):
            return self.collate_fn([self.__getitem__(ix) for ix in index])
        if isinstance(index, str):
            return self.__getitem__(self.file_names.index(index))
        return self.get_one_item(index)

    def __len__(self) -> int:
        return len(self.files_per_repr[self.task_names[0]]) # all of them have the same number (filled with None or not)

    def __str__(self):
        f_str = f"[{str(type(self)).rsplit('.', maxsplit=1)[-1][0:-2]}]"
        f_str += f"\n - Path: '{self.path}'"
        f_str += f"\n - Tasks ({len(self.tasks)}): {self.tasks}"
        f_str += f"\n - Classification tasks ({len(self.classification_tasks)}): {self.classification_tasks}"
        f_str += f"\n - Length: {len(self)}"
        f_str += f"\n - Handle missing data mode: '{self.handle_missing_data}'"
        f_str += f"\n - Normalization: '{self.normalization}'"
        return f_str

    def __repr__(self):
        return str(self)



# pylint: disable=all

def get_output_shape(oldh: int, oldw: int, short_edge_length: int, max_size: int):
    """Compute the output size given input size and target short edge length."""
    scale = short_edge_length / min(oldh, oldw)
    newh, neww = (short_edge_length, scale * oldw) if oldh < oldw else (scale * oldh, short_edge_length)
    if max(newh, neww) > max_size:
        scale = max_size * 1.0 / max(newh, neww)
        newh, neww = newh * scale, neww * scale
    neww, newh = int(neww + 0.5), int(newh + 0.5)
    return newh, neww



"""Canny edge detector representation."""
import numpy as np
from overrides import overrides

from vre.utils import MemoryData, ReprOut
from vre.vre_video import VREVideo
from vre.utils.cv2_utils import cv2_Canny

from ..edges import EdgesRepresentation

class Canny(EdgesRepresentation):
    """Canny edge detector representation."""
    def __init__(self, threshold1: float, threshold2: float, aperture_size: int, l2_gradient: bool, **kwargs):
        super().__init__(**kwargs)
        self.threshold1 = threshold1
        self.threshold2 = threshold2
        self.aperture_size = aperture_size
        self.l2_gradient = l2_gradient

    @overrides
    def compute(self, video: VREVideo, ixs: list[int]):
        assert self.data is None, f"[{self}] data must not be computed before calling this"
        self.data = ReprOut(frames=video[ixs], key=ixs,
                            output=MemoryData([self._make_one(frame) for frame in video[ixs]]))

    def _make_one(self, x: np.ndarray) -> np.ndarray:
        res = cv2_Canny(x, threshold1=self.threshold1, threshold2=self.threshold2,
                        apertureSize=self.aperture_size, L2gradient=self.l2_gradient)
        res = np.float32(res) / 255
        return res[..., None]



# pylint: disable=all
import numpy as np
import torch
from .utils import image_to_np_ndarray
from PIL import Image

from vre.utils.cv2_utils import (cv2_findContours, cv2_boundingRect, cv2_drawContours, cv2_RETR_TREE,
                                 cv2_CHAIN_APPROX_SIMPLE, cv2_RETR_EXTERNAL, cv2_COLOR_BGR2RGB, cv2_cvtColor)
from vre.utils import image_resize, array_blend

class FastSAMPrompt:
    def __init__(self, image, results, device='cuda'):
        if isinstance(image, str) or isinstance(image, Image.Image):
            image = image_to_np_ndarray(image)
        self.device = device
        self.results = results
        self.img = image

    def _segment_image(self, image, bbox):
        if isinstance(image, Image.Image):
            image_array = np.array(image)
        else:
            image_array = image
        segmented_image_array = np.zeros_like(image_array)
        x1, y1, x2, y2 = bbox
        segmented_image_array[y1:y2, x1:x2] = image_array[y1:y2, x1:x2]
        segmented_image = Image.fromarray(segmented_image_array)
        black_image = Image.new('RGB', image.size, (255, 255, 255))
        # transparency_mask = np.zeros_like((), dtype=np.uint8)
        transparency_mask = np.zeros((image_array.shape[0], image_array.shape[1]), dtype=np.uint8)
        transparency_mask[y1:y2, x1:x2] = 255
        transparency_mask_image = Image.fromarray(transparency_mask, mode='L')
        black_image.paste(segmented_image, mask=transparency_mask_image)
        return black_image

    def _format_results(self, result, filter=0):
        annotations = []
        n = len(result.masks.data)
        for i in range(n):
            annotation = {}
            mask = result.masks.data[i] == 1.0

            if torch.sum(mask) < filter:
                continue
            annotation['id'] = i
            annotation['segmentation'] = mask.cpu().numpy()
            annotation['bbox'] = result.boxes.data[i]
            annotation['score'] = result.boxes.conf[i]
            annotation['area'] = annotation['segmentation'].sum()
            annotations.append(annotation)
        return annotations

    def filter_masks(annotations):  # filte the overlap mask
        annotations.sort(key=lambda x: x['area'], reverse=True)
        to_remove = set()
        for i in range(0, len(annotations)):
            a = annotations[i]
            for j in range(i + 1, len(annotations)):
                b = annotations[j]
                if i != j and j not in to_remove:
                    # check if
                    if b['area'] < a['area']:
                        if (a['segmentation'] & b['segmentation']).sum() / b['segmentation'].sum() > 0.8:
                            to_remove.add(j)

        return [a for i, a in enumerate(annotations) if i not in to_remove], to_remove

    def _get_bbox_from_mask(self, mask):
        mask = mask.astype(np.uint8)
        contours, _ = cv2_findContours(mask, cv2_RETR_EXTERNAL, cv2_CHAIN_APPROX_SIMPLE)
        x1, y1, w, h = cv2_boundingRect(contours[0])
        x2, y2 = x1 + w, y1 + h
        if len(contours) > 1:
            for b in contours:
                x_t, y_t, w_t, h_t = cv2_boundingRect(b)
                # Merge multiple bounding boxes into one.
                x1 = min(x1, x_t)
                y1 = min(y1, y_t)
                x2 = max(x2, x_t + w_t)
                y2 = max(y2, y_t + h_t)
            h = y2 - y1
            w = x2 - x1
        return [x1, y1, x2, y2]

    def plot_to_result(self,
             annotations: torch.Tensor,
             mask_random_color=True,
             retina=False,
             withContours=True) -> np.ndarray:
        if isinstance(annotations[0], dict):
            annotations = [annotation['segmentation'] for annotation in annotations]

        image = self.img.astype(float) / 255
        original_h, original_w = image.shape[0:2]

        mask: np.ndarray = self.fast_show_mask(
            annotations,
            random_color=mask_random_color,
            retinamask=retina,
            target_height=original_h,
            target_width=original_w,
        )
        res = array_blend(image, mask[..., 0:3], mask[..., 3:4])

        if withContours:
            contour_all = []
            temp = np.zeros((original_h, original_w, 1))
            annotations_np = annotations.cpu().numpy()
            for mask in annotations_np:
                if type(mask) == dict:
                    mask = mask['segmentation']
                annotation = mask.astype(np.uint8)
                if not retina:
                    annotation = image_resize(annotation, original_h, original_w, "nearest", library="cv2")
                contours, _ = cv2_findContours(annotation, cv2_RETR_TREE, cv2_CHAIN_APPROX_SIMPLE)
                for contour in contours:
                    contour_all.append(contour)
            cv2_drawContours(temp, contour_all, -1, (255, 255, 255), 2)
            color = np.array([0 / 255, 0 / 255, 255 / 255, 0.8])
            contour_mask = temp / 255 * color.reshape(1, 1, -1)
            res = array_blend(res, contour_mask[..., 0:3], contour_mask[..., 3:4])
        return (res * 255).astype(np.uint8)

    def fast_show_mask(
        self,
        annotation,
        random_color=False,
        retinamask=True,
        target_height=960,
        target_width=960,
    ) -> np.ndarray:
        msak_sum = annotation.shape[0]
        height = annotation.shape[1]
        weight = annotation.shape[2]
        areas = torch.sum(annotation, dim=(1, 2))
        sorted_indices = torch.argsort(areas, descending=False)
        annotation = annotation[sorted_indices]
        # Find the index of the first non-zero value at each position.
        index = (annotation != 0).to(torch.long).argmax(dim=0)
        if random_color:
            color = torch.rand((msak_sum, 1, 1, 3)).to(annotation.device)
        else:
            color = torch.ones((msak_sum, 1, 1, 3)).to(annotation.device) * torch.tensor([
                30 / 255, 144 / 255, 255 / 255]).to(annotation.device)
        transparency = torch.ones((msak_sum, 1, 1, 1)).to(annotation.device) * 0.6
        visual = torch.cat([color, transparency], dim=-1)
        mask_image = torch.unsqueeze(annotation, -1) * visual
        # Select data according to the index. The index indicates which batch's data to choose at each position, converting the mask_image into a single batch form.
        show = torch.zeros((height, weight, 4)).to(annotation.device)
        h_indices, w_indices = torch.meshgrid(torch.arange(height), torch.arange(weight), indexing='ij')
        indices = (index[h_indices, w_indices], h_indices, w_indices, slice(None))
        # Use vectorized indexing to update the values of 'show'.
        show[h_indices, w_indices, :] = mask_image[indices]
        show_cpu = show.cpu().numpy()
        if not retinamask:
            show_cpu = image_resize(show_cpu, target_height, target_width, "nearest", library="cv2")
        return show_cpu

    # clip
    @torch.no_grad()
    def retrieve(self, model, preprocess, elements, search_text: str, device) -> int:
        import clip
        preprocessed_images = [preprocess(image).to(device) for image in elements]
        tokenized_text = clip.tokenize([search_text]).to(device)
        stacked_images = torch.stack(preprocessed_images)
        image_features = model.encode_image(stacked_images)
        text_features = model.encode_text(tokenized_text)
        image_features /= image_features.norm(dim=-1, keepdim=True)
        text_features /= text_features.norm(dim=-1, keepdim=True)
        probs = 100.0 * image_features @ text_features.T
        return probs[:, 0].softmax(dim=0)

    def _crop_image(self, format_results):

        image = Image.fromarray(cv2_cvtColor(self.img, cv2_COLOR_BGR2RGB))
        ori_w, ori_h = image.size
        annotations = format_results
        mask_h, mask_w = annotations[0]['segmentation'].shape
        if ori_w != mask_w or ori_h != mask_h:
            image = image.resize((mask_w, mask_h))
        cropped_boxes = []
        cropped_images = []
        not_crop = []
        filter_id = []
        # annotations, _ = filter_masks(annotations)
        # filter_id = list(_)
        for _, mask in enumerate(annotations):
            if np.sum(mask['segmentation']) <= 100:
                filter_id.append(_)
                continue
            bbox = self._get_bbox_from_mask(mask['segmentation'])  # mask 的 bbox
            cropped_boxes.append(self._segment_image(image, bbox))
            # cropped_boxes.append(segment_image(image,mask["segmentation"]))
            cropped_images.append(bbox)  # Save the bounding box of the cropped image.

        return cropped_boxes, cropped_images, not_crop, filter_id, annotations

    def box_prompt(self, bbox=None, bboxes=None):
        if self.results == None:
            return []
        assert bbox or bboxes
        if bboxes is None:
            bboxes = [bbox]
        max_iou_index = []
        for bbox in bboxes:
            assert (bbox[2] != 0 and bbox[3] != 0)
            masks = self.results[0].masks.data
            target_height = self.img.shape[0]
            target_width = self.img.shape[1]
            h = masks.shape[1]
            w = masks.shape[2]
            if h != target_height or w != target_width:
                bbox = [
                    int(bbox[0] * w / target_width),
                    int(bbox[1] * h / target_height),
                    int(bbox[2] * w / target_width),
                    int(bbox[3] * h / target_height), ]
            bbox[0] = round(bbox[0]) if round(bbox[0]) > 0 else 0
            bbox[1] = round(bbox[1]) if round(bbox[1]) > 0 else 0
            bbox[2] = round(bbox[2]) if round(bbox[2]) < w else w
            bbox[3] = round(bbox[3]) if round(bbox[3]) < h else h

            # IoUs = torch.zeros(len(masks), dtype=torch.float32)
            bbox_area = (bbox[3] - bbox[1]) * (bbox[2] - bbox[0])

            masks_area = torch.sum(masks[:, bbox[1]:bbox[3], bbox[0]:bbox[2]], dim=(1, 2))
            orig_masks_area = torch.sum(masks, dim=(1, 2))

            union = bbox_area + orig_masks_area - masks_area
            IoUs = masks_area / union
            max_iou_index.append(int(torch.argmax(IoUs)))
        max_iou_index = list(set(max_iou_index))
        return np.array(masks[max_iou_index].cpu().numpy())

    def point_prompt(self, points, pointlabel):  # numpy
        if self.results == None:
            return []
        masks = self._format_results(self.results[0], 0)
        target_height = self.img.shape[0]
        target_width = self.img.shape[1]
        h = masks[0]['segmentation'].shape[0]
        w = masks[0]['segmentation'].shape[1]
        if h != target_height or w != target_width:
            points = [[int(point[0] * w / target_width), int(point[1] * h / target_height)] for point in points]
        onemask = np.zeros((h, w))
        masks = sorted(masks, key=lambda x: x['area'], reverse=True)
        for i, annotation in enumerate(masks):
            if type(annotation) == dict:
                mask = annotation['segmentation']
            else:
                mask = annotation
            for i, point in enumerate(points):
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 1:
                    onemask[mask] = 1
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 0:
                    onemask[mask] = 0
        onemask = onemask >= 1
        return np.array([onemask])

    def text_prompt(self, text):
        import clip
        if self.results == None:
            return []
        format_results = self._format_results(self.results[0], 0)
        cropped_boxes, cropped_images, not_crop, filter_id, annotations = self._crop_image(format_results)
        clip_model, preprocess = clip.load('ViT-B/32', device=self.device)
        scores = self.retrieve(clip_model, preprocess, cropped_boxes, text, device=self.device)
        max_idx = scores.argsort()
        max_idx = max_idx[-1]
        max_idx += sum(np.array(filter_id) <= int(max_idx))
        return np.array([annotations[max_idx]['segmentation']])

    def everything_prompt(self):
        if self.results == None:
            return []
        return self.results[0].masks.data




"""Init file"""
# pylint: disable=reimported, wrong-import-position
import warnings
import sys
from pathlib import Path
from diffusers.utils.logging import disable_progress_bar


from .video_representations_extractor import VideoRepresentationsExtractor, VideoRepresentationsExtractor as VRE
from .representations import Representation, ReprOut
from .vre_video import VREVideo, FFmpegVideo, FakeVideo
from .utils.lovely import monkey_patch
from .utils.repr_memory_layout import MemoryData, DiskData

try:
    import pdbp
except ImportError:
    pass

monkey_patch()
disable_progress_bar()

warnings.filterwarnings("ignore", "You are using `torch.load` with `weights_only=False`*.")
warnings.filterwarnings("ignore", "`clean_up_tokenization_spaces` was not set*.")



"""Video Repentations Extractor module"""
from __future__ import annotations
from pathlib import Path
from datetime import datetime
import traceback
from tqdm import tqdm
import torch as tr
import numpy as np

from .representations import Representation, ComputeRepresentationMixin, LearnedRepresentationMixin
from .representations.io_representation_mixin import IORepresentationMixin
from .vre_runtime_args import VRERuntimeArgs
from .data_writer import DataWriter
from .data_storer import DataStorer
from .metadata import Metadata
from .vre_video import VREVideo
from .utils import now_fmt, make_batches, vre_topo_sort, ReprOut, DiskData
from .logger import vre_logger as logger

class VideoRepresentationsExtractor:
    """Video Representations Extractor class"""

    def __init__(self, video: VREVideo, representations: list[Representation]):
        """
        Parameters:
        - video The video we are performing VRE one
        - representations The list instantiated representations. Must be topo-sortable based on name and deps.
        """
        assert len(representations) > 0, "At least one representation must be provided"
        assert all(isinstance(x, Representation) for x in representations), [(x.name, type(x)) for x in representations]
        assert isinstance(video, VREVideo), (type(video), video)
        self.video = video
        self.representations: list[Representation] = vre_topo_sort(representations)
        self.repr_names = [r.name for r in representations]
        self._logs_file: Path | None = None
        self._metadata: Metadata | None = None

    def set_compute_params(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations of ComputeRepresentationMixin type"""
        for r in [_r for _r in self.representations if isinstance(_r, ComputeRepresentationMixin)]:
            r.set_compute_params(**kwargs)
        return self

    def set_io_parameters(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations of IORepresentationMixin type"""
        for r in [_r for _r in self.representations if isinstance(_r, IORepresentationMixin)]:
            r.set_io_params(**kwargs)
        return self

    def to_graphviz(self, **kwargs) -> "Digraph":
        """Returns a graphviz object from this graph. Used for plotting the graph. Best for smaller graphs."""
        from graphviz import Digraph # pylint: disable=import-outside-toplevel
        g = Digraph()
        for k, v in kwargs.items():
            g.attr(**{k: v})
        g.attr(rankdir="LR")
        for node in self.representations:
            g.node(name=f"{node.name}", shape="oval")
        edges: list[tuple[str, str]] = [(r.name, dep.name) for r in self.representations for dep in r.dependencies]
        for l, r in edges:
            g.edge(r, l) # reverse?
        return g

    def run(self, output_dir: Path, frames: list[int] | None = None, output_dir_exists_mode: str = "raise",
            exception_mode: str = "stop_execution", n_threads_data_storer: int = 0,
            exported_representations: list[str] | None = None) -> Metadata:
        """
        The main loop of the VRE. This will run all the representations on the video and store results in the output_dir
        Parameters:
        - output_dir The directory where VRE will store the representations
        - frames The list of frames to run VRE for. If None, will export all the frames of the video
        - exception_mode What to do when encountering an exception. It always writes the exception to 'exception.txt'
          - 'skip_representation' Will stop the run of the current representation and start the next one
          - 'stop_execution' (default) Will stop the execution of VRE
        - n_threads_data_storer The number of threads used by the DataStorer
        - exported_representations If set, only this subset of representations are exported, otherwise all of them
            based on these provided to the VRE constructor.
        Returns:
        - A dataframe with the run statistics for each representation
        """
        self._setup_logger(output_dir, now := now_fmt())
        try:
            self.to_graphviz().render(pth := f"{output_dir}/.logs/graph-{now}", format="png", cleanup=True)
            logger.info(f"Stored graphviz representation at: '{pth}.png'")
        except ImportError:
            pass

        runtime_args = VRERuntimeArgs(self.video, self.representations, frames, exception_mode, n_threads_data_storer)
        logger.info(runtime_args)
        self._metadata = Metadata(self.repr_names, runtime_args, self._logs_file.parent / f"run_metadata-{now}.json")
        for vre_repr in self._get_output_representations(exported_representations):
            assert isinstance(vre_repr, Representation), vre_repr
            assert isinstance(vre_repr, IORepresentationMixin), vre_repr
            assert vre_repr.binary_format is not None or vre_repr.image_format is not None, vre_repr
            data_writer = DataWriter(output_dir, representation=vre_repr, output_dir_exists_mode=output_dir_exists_mode)
            repr_had_exception = self._do_one_representation(data_writer, runtime_args) # vre_repr is part of writer
            if repr_had_exception and runtime_args.exception_mode == "stop_execution":
                raise RuntimeError(f"Representation '{vre_repr.name}' threw. Check '{self._logs_file}' for information")
            vre_repr.vre_free() if isinstance(vre_repr, LearnedRepresentationMixin) and vre_repr.setup_called else None
        self._end_run()
        return self._metadata

    # Private methods related to the main computation of each representation
    def _cleanup_one_representation(self, representation: Representation, data_storer: DataStorer):
        data_storer.join_with_timeout(timeout=30)
        if isinstance(representation, LearnedRepresentationMixin) and representation.setup_called:
            representation.vre_free()
        for dep in representation.dependencies:
            if isinstance(dep, LearnedRepresentationMixin) and dep.setup_called:
                dep.vre_free()

    def _load_from_disk_if_possible(self, rep: Representation, video: VREVideo, ixs: list[int], output_dir: Path):
        """loads (batched) data from disk if possible."""
        assert isinstance(rep, IORepresentationMixin), rep
        assert isinstance(ixs, list) and all(isinstance(ix, int) for ix in ixs), (type(ixs), [type(ix) for ix in ixs])
        assert output_dir is not None and output_dir.exists(), output_dir
        # TODO: use data writer to create the paths as they can be non-npz for example
        npz_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}.npz" for ix in ixs]
        extra_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}_extra.npz" for ix in ixs]
        if any(not x.exists() for x in npz_paths): # partial batches are considered 'not existing' and overwritten
            return
        extras_exist = [x.exists() for x in extra_paths]
        assert (ee := sum(extras_exist)) in (0, (ep := len(extra_paths))), f"Found {ee}. Expected either 0 or {ep}"
        disk_data: DiskData = np.array([rep.load_from_disk(x) for x in npz_paths])
        extra = [np.load(x, allow_pickle=True)["arr_0"].item() for x in extra_paths] if ee == ep else None
        logger.debug2(f"[{rep}] Slice: [{ixs[0]}:{ixs[-1]}]. All data found on disk and loaded")
        rep.data = ReprOut(frames=video[ixs], output=rep.disk_to_memory_fmt(disk_data), extra=extra, key=ixs)

    def _compute_one_representation_batch(self, rep: Representation, batch: list[int],
                                          output_dir: Path, depth: int = 0):
        assert isinstance(rep, ComputeRepresentationMixin), rep
        # setting depth to higher values allows to compute deps of deps in memory. We throw to not hide bugs.
        # useful for VRE hacking (i.e. see semanit_mapper.py)
        assert depth <= 1, f"{rep=} {depth=} {batch=} {output_dir=}"
        # TODO: make unit test with this (lingering .data from previous computation) on depth == 1 (i.e. deps of rep)
        rep.data = None # Important to invalidate any previous results here
        self._load_from_disk_if_possible(rep, self.video, batch, output_dir)
        if rep.data is not None:
            return
        for dep in rep.dependencies:
            self._compute_one_representation_batch(dep, batch, output_dir, depth + 1)
        if isinstance(rep, LearnedRepresentationMixin) and rep.setup_called is False:
            rep.vre_setup() # instantiates the model, loads to cuda device etc.
        rep.compute(self.video, ixs=batch)
        assert rep.data is not None, f"{rep=} {batch=} {output_dir=} {depth=}"

    def _do_one_representation(self, data_writer: DataWriter, runtime_args: VRERuntimeArgs) -> bool:
        """main loop of each representation. Returns true if had exception, false otherwise"""
        data_storer = DataStorer(data_writer, runtime_args.n_threads_data_storer)
        logger.info(f"Running:\n{data_writer.rep}\n{data_storer}")
        rep: Representation | ComputeRepresentationMixin | IORepresentationMixin = data_writer.rep
        self._metadata.metadata["data_writers"][rep.name] = data_writer.to_dict()
        rep.output_size = self.video.frame_shape[0:2] if rep.output_size == "video_shape" else rep.output_size

        batches = make_batches(runtime_args.frames, rep.batch_size)
        pbar = tqdm(total=runtime_args.n_frames, desc=f"[VRE] {rep.name} bs={rep.batch_size}")
        for i, batch in enumerate(batches):
            if i % runtime_args.store_metadata_every_n_iters == 0:
                self._metadata.store_on_disk()

            now = datetime.now()
            if data_writer.all_batch_exists(batch):
                self._metadata.add_time(rep.name, (datetime.now() - now).total_seconds(), len(batch))
                pbar.update(len(batch))
                continue

            try:
                tr.cuda.empty_cache() # might empty some unused memory, not 100% if needed.
                self._compute_one_representation_batch(rep=rep, batch=batch, output_dir=data_writer.output_dir)
                assert rep.data is not None, (rep, batch)
                if not rep.is_classification and rep.name.find("fastsam") == -1: # TODO: MemoryData of M2F is binary...
                    assert rep.data.output.shape[-1] == rep.n_channels, (rep, rep.data.output, rep.n_channels)
                rep.data.output_images = rep.make_images(rep.data) if rep.export_image else None
                data_storer(rep.data)
            except Exception:
                self._log_error(f"\n[{rep.name} {rep.batch_size=} {batch=}] {traceback.format_exc()}\n")
                self._metadata.add_time(rep.name, 1 << 31, len(runtime_args.frames) - i * rep.batch_size)
                self._cleanup_one_representation(rep, data_storer)
                return True
            self._metadata.add_time(rep.name, (datetime.now() - now).total_seconds(), len(batch))
            pbar.update(len(batch))
        self._cleanup_one_representation(rep, data_storer)
        return False

    # Helper private methods
    def _setup_logger(self, output_dir: Path, now_str: str):
        (logs_dir := output_dir / ".logs").mkdir(exist_ok=True, parents=True)
        self._logs_file = logs_dir / f"logs-{now_str}.txt"
        logger.add_file_handler(self._logs_file)

    def _log_error(self, msg: str):
        assert self._logs_file is not None, "_log_error can be called only after run() when self._log_file is set"
        self._logs_file.parent.mkdir(exist_ok=True, parents=True)
        open(self._logs_file, "a").write(f"{'=' * 80}\n{now_fmt()}\n{msg}\n{'=' * 80}")
        logger.debug(f"Error: {msg}")

    def _end_run(self):
        logger.remove_file_handler()
        self._metadata.store_on_disk()
        logger.info(f"Stored vre run log file at '{self._metadata.disk_location}")

    def _get_output_representations(self, exported_representations: list[str]) -> list[IORepresentationMixin]:
        """given all the representations, keep those that actually export something. At least one is needed"""
        crs: list[IORepresentationMixin] = [_r for _r in self.representations if isinstance(_r, IORepresentationMixin)]
        assert len(crs) > 0, f"No I/O Representation found in {self.repr_names}"
        out_r: list[Representation] = [r for r in crs if r.export_binary or r.export_image]
        if exported_representations:
            logger.info(f"Explicit subset providewd: {exported_representations}. Exporting only these.")
            out_r = [r for r in out_r if r.name in exported_representations]
        assert len(out_r) > 0, f"No output format set for any I/O Representation: {', '.join([r.name for r in crs])}"
        return out_r

    def __call__(self, *args, **kwargs) -> Metadata:
        return self.run(*args, **kwargs)

    def __str__(self) -> str:
        return f"""\n[VRE]
- Video: {self.video}
- Representations ({len(self.representations)}): [{", ".join(self.repr_names)}]"""

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, key: str) -> Representation:
        assert key in self.repr_names, f"Representation '{key}' not in {self.repr_names}"
        return self.representations[self.repr_names.index(key)]



"""Init file"""
# pylint: disable=reimported

from .representation import Representation, ReprOut
from .task_mapper import TaskMapper
from .build_representations import add_external_repositories, build_representations_from_cfg

from .learned_representation_mixin import LearnedRepresentationMixin
from .compute_representation_mixin import ComputeRepresentationMixin
from .normed_representation_mixin import NormedRepresentationMixin
from .io_representation_mixin import IORepresentationMixin
from .np_io_representation import NpIORepresentation



"""init file for vre default repository"""
from typing import Type
from vre.representations import Representation

# pylint: disable=import-outside-toplevel
def get_vre_repository() -> dict[str, Type[Representation]]:
    """the built-in repository for VRE"""
    from .color.rgb import RGB
    from .color.hsv import HSV
    from .soft_segmentation.halftone import Halftone
    from .soft_segmentation.generalized_boundaries import GeneralizedBoundaries
    from .soft_segmentation.fastsam import FastSam
    from .edges.canny import Canny
    from .edges.dexined import DexiNed
    from .depth.dpt import DepthDpt
    from .depth.marigold import Marigold
    from .optical_flow.rife import FlowRife
    from .optical_flow.raft import FlowRaft
    from .semantic_segmentation.safeuav import SafeUAV
    from .semantic_segmentation.mask2former import Mask2Former
    from .normals.depth_svd import DepthNormalsSVD

    return {
        "color/rgb": RGB,
        "color/hsv": HSV,
        "soft-segmentation/python-halftone": Halftone,
        "soft-segmentation/generalized-boundaries": GeneralizedBoundaries,
        "soft-segmentation/fastsam": FastSam,
        "edges/canny": Canny,
        "edges/dexined": DexiNed,
        "depth/dpt": DepthDpt,
        "depth/marigold": Marigold,
        "semantic-segmentation/safeuav": SafeUAV,
        "semantic-segmentation/mask2former": Mask2Former,
        "optical-flow/rife": FlowRife,
        "optical-flow/raft": FlowRaft,
        "normals/depth-svd": DepthNormalsSVD,
    }



"""YACS -- Yet Another Configuration System is designed to be a simple
configuration management system for academic and industrial research
projects.

See README.md for usage and examples.
"""
# pylint: disable=all
import copy
import io
import logging
import os
import sys
from ast import literal_eval

import yaml

# Flag for py2 and py3 compatibility to use when separate code paths are necessary
# When _PY2 is False, we assume Python 3 is in use

# Filename extensions for loading configs from files
_YAML_EXTS = {"", ".yaml", ".yml"}
_PY_EXTS = {".py"}

# py2 and py3 compatibility for checking file object type
# We simply use this to infer py2 vs py3
_FILE_TYPES = (io.IOBase,)

# CfgNodes can only contain a limited set of valid types
_VALID_TYPES = {tuple, list, str, int, float, bool, type(None)}
# py2 allow for str and unicode

# Utilities for importing modules from file paths
import importlib.util

logger = logging.getLogger(__name__)


class CfgNode(dict):
    """
    CfgNode represents an internal node in the configuration tree. It's a simple
    dict-like container that allows for attribute-based access to keys.
    """

    IMMUTABLE = "__immutable__"
    DEPRECATED_KEYS = "__deprecated_keys__"
    RENAMED_KEYS = "__renamed_keys__"
    NEW_ALLOWED = "__new_allowed__"

    def __init__(self, init_dict=None, key_list=None, new_allowed=False):
        """
        Args:
            init_dict (dict): the possibly-nested dictionary to initailize the CfgNode.
            key_list (list[str]): a list of names which index this CfgNode from the root.
                Currently only used for logging purposes.
            new_allowed (bool): whether adding new key is allowed when merging with
                other configs.
        """
        # Recursively convert nested dictionaries in init_dict into CfgNodes
        init_dict = {} if init_dict is None else init_dict
        key_list = [] if key_list is None else key_list
        init_dict = self._create_config_tree_from_dict(init_dict, key_list)
        super(CfgNode, self).__init__(init_dict)
        # Manage if the CfgNode is frozen or not
        self.__dict__[CfgNode.IMMUTABLE] = False
        # Deprecated options
        # If an option is removed from the code and you don't want to break existing
        # yaml configs, you can add the full config key as a string to the set below.
        self.__dict__[CfgNode.DEPRECATED_KEYS] = set()
        # Renamed options
        # If you rename a config option, record the mapping from the old name to the new
        # name in the dictionary below. Optionally, if the type also changed, you can
        # make the value a tuple that specifies first the renamed key and then
        # instructions for how to edit the config file.
        self.__dict__[CfgNode.RENAMED_KEYS] = {
            # 'EXAMPLE.OLD.KEY': 'EXAMPLE.NEW.KEY',  # Dummy example to follow
            # 'EXAMPLE.OLD.KEY': (                   # A more complex example to follow
            #     'EXAMPLE.NEW.KEY',
            #     "Also convert to a tuple, e.g., 'foo' -> ('foo',) or "
            #     + "'foo:bar' -> ('foo', 'bar')"
            # ),
        }

        # Allow new attributes after initialisation
        self.__dict__[CfgNode.NEW_ALLOWED] = new_allowed

    @classmethod
    def _create_config_tree_from_dict(cls, dic, key_list):
        """
        Create a configuration tree using the given dict.
        Any dict-like objects inside dict will be treated as a new CfgNode.

        Args:
            dic (dict):
            key_list (list[str]): a list of names which index this CfgNode from the root.
                Currently only used for logging purposes.
        """
        dic = copy.deepcopy(dic)
        for k, v in dic.items():
            if isinstance(v, dict):
                # Convert dict to CfgNode
                dic[k] = cls(v, key_list=key_list + [k])
            else:
                # Check for valid leaf type or nested CfgNode
                _assert_with_logging(
                    _valid_type(v, allow_cfg_node=False),
                    "Key {} with value {} is not a valid type; valid types: {}".format(
                        ".".join(key_list + [str(k)]), type(v), _VALID_TYPES
                    ),
                )
        return dic

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self.is_frozen():
            raise AttributeError(
                "Attempted to set {} to {}, but CfgNode is immutable".format(
                    name, value
                )
            )

        _assert_with_logging(
            name not in self.__dict__,
            "Invalid attempt to modify internal CfgNode state: {}".format(name),
        )
        _assert_with_logging(
            _valid_type(value, allow_cfg_node=True),
            "Invalid type {} for key {}; valid types = {}".format(
                type(value), name, _VALID_TYPES
            ),
        )

        self[name] = value

    def __str__(self):
        def _indent(s_, num_spaces):
            s = s_.split("\n")
            if len(s) == 1:
                return s_
            first = s.pop(0)
            s = [(num_spaces * " ") + line for line in s]
            s = "\n".join(s)
            s = first + "\n" + s
            return s

        r = ""
        s = []
        for k, v in sorted(self.items()):
            seperator = "\n" if isinstance(v, CfgNode) else " "
            attr_str = "{}:{}{}".format(str(k), seperator, str(v))
            attr_str = _indent(attr_str, 2)
            s.append(attr_str)
        r += "\n".join(s)
        return r

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, super(CfgNode, self).__repr__())

    def dump(self, **kwargs):
        """Dump to a string."""

        def convert_to_dict(cfg_node, key_list):
            if not isinstance(cfg_node, CfgNode):
                _assert_with_logging(
                    _valid_type(cfg_node),
                    "Key {} with value {} is not a valid type; valid types: {}".format(
                        ".".join(key_list), type(cfg_node), _VALID_TYPES
                    ),
                )
                return cfg_node
            else:
                cfg_dict = dict(cfg_node)
                for k, v in cfg_dict.items():
                    cfg_dict[k] = convert_to_dict(v, key_list + [k])
                return cfg_dict

        self_as_dict = convert_to_dict(self, [])
        return yaml.safe_dump(self_as_dict, **kwargs)

    def merge_from_file(self, cfg_filename):
        """Load a yaml config file and merge it this CfgNode."""
        with open(cfg_filename, "r") as f:
            cfg = self.load_cfg(f)
        self.merge_from_other_cfg(cfg)

    def merge_from_other_cfg(self, cfg_other):
        """Merge `cfg_other` into this CfgNode."""
        _merge_a_into_b(cfg_other, self, self, [])

    def merge_from_list(self, cfg_list):
        """Merge config (keys, values) in a list (e.g., from command line) into
        this CfgNode. For example, `cfg_list = ['FOO.BAR', 0.5]`.
        """
        _assert_with_logging(
            len(cfg_list) % 2 == 0,
            "Override list has odd length: {}; it must be a list of pairs".format(
                cfg_list
            ),
        )
        root = self
        for full_key, v in zip(cfg_list[0::2], cfg_list[1::2]):
            if root.key_is_deprecated(full_key):
                continue
            if root.key_is_renamed(full_key):
                root.raise_key_rename_error(full_key)
            key_list = full_key.split(".")
            d = self
            for subkey in key_list[:-1]:
                _assert_with_logging(
                    subkey in d, "Non-existent key: {}".format(full_key)
                )
                d = d[subkey]
            subkey = key_list[-1]
            _assert_with_logging(subkey in d, "Non-existent key: {}".format(full_key))
            value = self._decode_cfg_value(v)
            value = _check_and_coerce_cfg_value_type(value, d[subkey], subkey, full_key)
            d[subkey] = value

    def freeze(self):
        """Make this CfgNode and all of its children immutable."""
        self._immutable(True)

    def defrost(self):
        """Make this CfgNode and all of its children mutable."""
        self._immutable(False)

    def is_frozen(self):
        """Return mutability."""
        return self.__dict__[CfgNode.IMMUTABLE]

    def _immutable(self, is_immutable):
        """Set immutability to is_immutable and recursively apply the setting
        to all nested CfgNodes.
        """
        self.__dict__[CfgNode.IMMUTABLE] = is_immutable
        # Recursively set immutable state
        for v in self.__dict__.values():
            if isinstance(v, CfgNode):
                v._immutable(is_immutable)
        for v in self.values():
            if isinstance(v, CfgNode):
                v._immutable(is_immutable)

    def clone(self):
        """Recursively copy this CfgNode."""
        return copy.deepcopy(self)

    def register_deprecated_key(self, key):
        """Register key (e.g. `FOO.BAR`) a deprecated option. When merging deprecated
        keys a warning is generated and the key is ignored.
        """
        _assert_with_logging(
            key not in self.__dict__[CfgNode.DEPRECATED_KEYS],
            "key {} is already registered as a deprecated key".format(key),
        )
        self.__dict__[CfgNode.DEPRECATED_KEYS].add(key)

    def register_renamed_key(self, old_name, new_name, message=None):
        """Register a key as having been renamed from `old_name` to `new_name`.
        When merging a renamed key, an exception is thrown alerting to user to
        the fact that the key has been renamed.
        """
        _assert_with_logging(
            old_name not in self.__dict__[CfgNode.RENAMED_KEYS],
            "key {} is already registered as a renamed cfg key".format(old_name),
        )
        value = new_name
        if message:
            value = (new_name, message)
        self.__dict__[CfgNode.RENAMED_KEYS][old_name] = value

    def key_is_deprecated(self, full_key):
        """Test if a key is deprecated."""
        if full_key in self.__dict__[CfgNode.DEPRECATED_KEYS]:
            logger.warning("Deprecated config key (ignoring): {}".format(full_key))
            return True
        return False

    def key_is_renamed(self, full_key):
        """Test if a key is renamed."""
        return full_key in self.__dict__[CfgNode.RENAMED_KEYS]

    def raise_key_rename_error(self, full_key):
        new_key = self.__dict__[CfgNode.RENAMED_KEYS][full_key]
        if isinstance(new_key, tuple):
            msg = " Note: " + new_key[1]
            new_key = new_key[0]
        else:
            msg = ""
        raise KeyError(
            "Key {} was renamed to {}; please update your config.{}".format(
                full_key, new_key, msg
            )
        )

    def is_new_allowed(self):
        return self.__dict__[CfgNode.NEW_ALLOWED]

    def set_new_allowed(self, is_new_allowed):
        """
        Set this config (and recursively its subconfigs) to allow merging
        new keys from other configs.
        """
        self.__dict__[CfgNode.NEW_ALLOWED] = is_new_allowed
        # Recursively set new_allowed state
        for v in self.__dict__.values():
            if isinstance(v, CfgNode):
                v.set_new_allowed(is_new_allowed)
        for v in self.values():
            if isinstance(v, CfgNode):
                v.set_new_allowed(is_new_allowed)

    @classmethod
    def load_cfg(cls, cfg_file_obj_or_str):
        """
        Load a cfg.
        Args:
            cfg_file_obj_or_str (str or file):
                Supports loading from:
                - A file object backed by a YAML file
                - A file object backed by a Python source file that exports an attribute
                  "cfg" that is either a dict or a CfgNode
                - A string that can be parsed as valid YAML
        """
        _assert_with_logging(
            isinstance(cfg_file_obj_or_str, _FILE_TYPES + (str,)),
            "Expected first argument to be of type {} or {}, but it was {}".format(
                _FILE_TYPES, str, type(cfg_file_obj_or_str)
            ),
        )
        if isinstance(cfg_file_obj_or_str, str):
            return cls._load_cfg_from_yaml_str(cfg_file_obj_or_str)
        elif isinstance(cfg_file_obj_or_str, _FILE_TYPES):
            return cls._load_cfg_from_file(cfg_file_obj_or_str)
        else:
            raise NotImplementedError("Impossible to reach here (unless there's a bug)")

    @classmethod
    def _load_cfg_from_file(cls, file_obj):
        """Load a config from a YAML file or a Python source file."""
        _, file_extension = os.path.splitext(file_obj.name)
        if file_extension in _YAML_EXTS:
            return cls._load_cfg_from_yaml_str(file_obj.read())
        elif file_extension in _PY_EXTS:
            return cls._load_cfg_py_source(file_obj.name)
        else:
            raise Exception(
                "Attempt to load from an unsupported file type {}; "
                "only {} are supported".format(file_obj, _YAML_EXTS.union(_PY_EXTS))
            )

    @classmethod
    def _load_cfg_from_yaml_str(cls, str_obj):
        """Load a config from a YAML string encoding."""
        cfg_as_dict = yaml.safe_load(str_obj)
        return cls(cfg_as_dict)

    @classmethod
    def _load_cfg_py_source(cls, filename):
        """Load a config from a Python source file."""
        module = _load_module_from_file("yacs.config.override", filename)
        _assert_with_logging(
            hasattr(module, "cfg"),
            "Python module from file {} must have 'cfg' attr".format(filename),
        )
        VALID_ATTR_TYPES = {dict, CfgNode}
        _assert_with_logging(
            type(module.cfg) in VALID_ATTR_TYPES,
            "Imported module 'cfg' attr must be in {} but is {} instead".format(
                VALID_ATTR_TYPES, type(module.cfg)
            ),
        )
        return cls(module.cfg)

    @classmethod
    def _decode_cfg_value(cls, value):
        """
        Decodes a raw config value (e.g., from a yaml config files or command
        line argument) into a Python object.

        If the value is a dict, it will be interpreted as a new CfgNode.
        If the value is a str, it will be evaluated as literals.
        Otherwise it is returned as-is.
        """
        # Configs parsed from raw yaml will contain dictionary keys that need to be
        # converted to CfgNode objects
        if isinstance(value, dict):
            return cls(value)
        # All remaining processing is only applied to strings
        if not isinstance(value, str):
            return value
        # Try to interpret `value` as a:
        #   string, number, tuple, list, dict, boolean, or None
        try:
            value = literal_eval(value)
        # The following two excepts allow v to pass through when it represents a
        # string.
        #
        # Longer explanation:
        # The type of v is always a string (before calling literal_eval), but
        # sometimes it *represents* a string and other times a data structure, like
        # a list. In the case that v represents a string, what we got back from the
        # yaml parser is 'foo' *without quotes* (so, not '"foo"'). literal_eval is
        # ok with '"foo"', but will raise a ValueError if given 'foo'. In other
        # cases, like paths (v = 'foo/bar' and not v = '"foo/bar"'), literal_eval
        # will raise a SyntaxError.
        except ValueError:
            pass
        except SyntaxError:
            pass
        return value


load_cfg = (
    CfgNode.load_cfg
)  # keep this function in global scope for backward compatibility


def _valid_type(value, allow_cfg_node=False):
    return (type(value) in _VALID_TYPES) or (
        allow_cfg_node and isinstance(value, CfgNode)
    )


def _merge_a_into_b(a, b, root, key_list):
    """Merge config dictionary a into config dictionary b, clobbering the
    options in b whenever they are also specified in a.
    """
    _assert_with_logging(
        isinstance(a, CfgNode),
        "`a` (cur type {}) must be an instance of {}".format(type(a), CfgNode),
    )
    _assert_with_logging(
        isinstance(b, CfgNode),
        "`b` (cur type {}) must be an instance of {}".format(type(b), CfgNode),
    )

    for k, v_ in a.items():
        full_key = ".".join(key_list + [k])

        v = copy.deepcopy(v_)
        v = b._decode_cfg_value(v)

        if k in b:
            v = _check_and_coerce_cfg_value_type(v, b[k], k, full_key)
            # Recursively merge dicts
            if isinstance(v, CfgNode):
                try:
                    _merge_a_into_b(v, b[k], root, key_list + [k])
                except BaseException:
                    raise
            else:
                b[k] = v
        elif b.is_new_allowed():
            b[k] = v
        else:
            if root.key_is_deprecated(full_key):
                continue
            elif root.key_is_renamed(full_key):
                root.raise_key_rename_error(full_key)
            else:
                raise KeyError("Non-existent config key: {}".format(full_key))


def _check_and_coerce_cfg_value_type(replacement, original, key, full_key):
    """Checks that `replacement`, which is intended to replace `original` is of
    the right type. The type is correct if it matches exactly or is one of a few
    cases in which the type can be easily coerced.
    """
    original_type = type(original)
    replacement_type = type(replacement)

    # The types must match (with some exceptions)
    if replacement_type == original_type:
        return replacement

    # If either of them is None, allow type conversion to one of the valid types
    if (replacement_type == type(None) and original_type in _VALID_TYPES) or (
        original_type == type(None) and replacement_type in _VALID_TYPES
    ):
        return replacement

    # Cast replacement from from_type to to_type if the replacement and original
    # types match from_type and to_type
    def conditional_cast(from_type, to_type):
        if replacement_type == from_type and original_type == to_type:
            return True, to_type(replacement)
        else:
            return False, None

    # Conditionally casts
    # list <-> tuple
    casts = [(tuple, list), (list, tuple)]
    # For py2: allow converting from str (bytes) to a unicode string
    try:
        casts.append((str, unicode))  # noqa: F821
    except Exception:
        pass

    for (from_type, to_type) in casts:
        converted, converted_value = conditional_cast(from_type, to_type)
        if converted:
            return converted_value

    raise ValueError(
        "Type mismatch ({} vs. {}) with values ({} vs. {}) for config "
        "key: {}".format(
            original_type, replacement_type, original, replacement, full_key
        )
    )


def _assert_with_logging(cond, msg):
    if not cond:
        logger.debug(msg)
    assert cond, msg


def _load_module_from_file(name, filename):
    spec = importlib.util.spec_from_file_location(name, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



"""VREVideo module"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable

class VREVideo(Iterable, ABC):
    """VREVideo -- A generic wrapper on top of a Video container"""
    def __init__(self):
        self.write_process = None

    @property
    @abstractmethod
    def shape(self) -> tuple[int, int, int, int]:
        """Returns the (N, H, W, 3) tuple of the video"""

    @property
    @abstractmethod
    def fps(self) -> float:
        """The frame rate of the video"""

    @abstractmethod
    def write(self, out_path: Path, start_frame: int = 0, end_frame: int | None = None):
        """writes the video to the path. TODO: add params to send to the underlying write impl (ffmpeg/image_write)"""

    def __iter__(self):
        index = 0
        try:
            while True:
                yield self[index]
                index += 1
        except IndexError: # why?
            pass



# pylint: disable=all
# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from: https://github.com/facebookresearch/detr/blob/master/models/transformer.py
"""
Transformer class.

Copy-paste from torch.nn.Transformer with modifications:
    * positional encodings are passed in MHattention
    * extra LN at the end of encoder is removed
    * decoder returns a stack of activations from all decoding layers
"""
import copy
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor, nn


class Transformer(nn.Module):
    def __init__(
        self,
        d_model=512,
        nhead=8,
        num_encoder_layers=6,
        num_decoder_layers=6,
        dim_feedforward=2048,
        dropout=0.1,
        activation="relu",
        normalize_before=False,
        return_intermediate_dec=False,
    ):
        super().__init__()

        encoder_layer = TransformerEncoderLayer(
            d_model, nhead, dim_feedforward, dropout, activation, normalize_before
        )
        encoder_norm = nn.LayerNorm(d_model) if normalize_before else None
        self.encoder = TransformerEncoder(encoder_layer, num_encoder_layers, encoder_norm)

        decoder_layer = TransformerDecoderLayer(
            d_model, nhead, dim_feedforward, dropout, activation, normalize_before
        )
        decoder_norm = nn.LayerNorm(d_model)
        self.decoder = TransformerDecoder(
            decoder_layer,
            num_decoder_layers,
            decoder_norm,
            return_intermediate=return_intermediate_dec,
        )

        self._reset_parameters()

        self.d_model = d_model
        self.nhead = nhead

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(self, src, mask, query_embed, pos_embed):
        # flatten NxCxHxW to HWxNxC
        bs, c, h, w = src.shape
        src = src.flatten(2).permute(2, 0, 1)
        pos_embed = pos_embed.flatten(2).permute(2, 0, 1)
        query_embed = query_embed.unsqueeze(1).repeat(1, bs, 1)
        if mask is not None:
            mask = mask.flatten(1)

        tgt = torch.zeros_like(query_embed)
        memory = self.encoder(src, src_key_padding_mask=mask, pos=pos_embed)
        hs = self.decoder(
            tgt, memory, memory_key_padding_mask=mask, pos=pos_embed, query_pos=query_embed
        )
        return hs.transpose(1, 2), memory.permute(1, 2, 0).view(bs, c, h, w)


class TransformerEncoder(nn.Module):
    def __init__(self, encoder_layer, num_layers, norm=None):
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src,
        mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
    ):
        output = src

        for layer in self.layers:
            output = layer(
                output, src_mask=mask, src_key_padding_mask=src_key_padding_mask, pos=pos
            )

        if self.norm is not None:
            output = self.norm(output)

        return output


class TransformerDecoder(nn.Module):
    def __init__(self, decoder_layer, num_layers, norm=None, return_intermediate=False):
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        self.return_intermediate = return_intermediate

    def forward(
        self,
        tgt,
        memory,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
        query_pos: Optional[Tensor] = None,
    ):
        output = tgt

        intermediate = []

        for layer in self.layers:
            output = layer(
                output,
                memory,
                tgt_mask=tgt_mask,
                memory_mask=memory_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                memory_key_padding_mask=memory_key_padding_mask,
                pos=pos,
                query_pos=query_pos,
            )
            if self.return_intermediate:
                intermediate.append(self.norm(output))

        if self.norm is not None:
            output = self.norm(output)
            if self.return_intermediate:
                intermediate.pop()
                intermediate.append(output)

        if self.return_intermediate:
            return torch.stack(intermediate)

        return output.unsqueeze(0)


class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model,
        nhead,
        dim_feedforward=2048,
        dropout=0.1,
        activation="relu",
        normalize_before=False,
    ):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(
        self,
        src,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
    ):
        q = k = self.with_pos_embed(src, pos)
        src2 = self.self_attn(
            q, k, value=src, attn_mask=src_mask, key_padding_mask=src_key_padding_mask
        )[0]
        src = src + self.dropout1(src2)
        src = self.norm1(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)
        return src

    def forward_pre(
        self,
        src,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
    ):
        src2 = self.norm1(src)
        q = k = self.with_pos_embed(src2, pos)
        src2 = self.self_attn(
            q, k, value=src2, attn_mask=src_mask, key_padding_mask=src_key_padding_mask
        )[0]
        src = src + self.dropout1(src2)
        src2 = self.norm2(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src2))))
        src = src + self.dropout2(src2)
        return src

    def forward(
        self,
        src,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
    ):
        if self.normalize_before:
            return self.forward_pre(src, src_mask, src_key_padding_mask, pos)
        return self.forward_post(src, src_mask, src_key_padding_mask, pos)


class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model,
        nhead,
        dim_feedforward=2048,
        dropout=0.1,
        activation="relu",
        normalize_before=False,
    ):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        self.multihead_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(
        self,
        tgt,
        memory,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
        query_pos: Optional[Tensor] = None,
    ):
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(
            q, k, value=tgt, attn_mask=tgt_mask, key_padding_mask=tgt_key_padding_mask
        )[0]
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)
        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
        )[0]
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout3(tgt2)
        tgt = self.norm3(tgt)
        return tgt

    def forward_pre(
        self,
        tgt,
        memory,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
        query_pos: Optional[Tensor] = None,
    ):
        tgt2 = self.norm1(tgt)
        q = k = self.with_pos_embed(tgt2, query_pos)
        tgt2 = self.self_attn(
            q, k, value=tgt2, attn_mask=tgt_mask, key_padding_mask=tgt_key_padding_mask
        )[0]
        tgt = tgt + self.dropout1(tgt2)
        tgt2 = self.norm2(tgt)
        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt2, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask,
        )[0]
        tgt = tgt + self.dropout2(tgt2)
        tgt2 = self.norm3(tgt)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt2))))
        tgt = tgt + self.dropout3(tgt2)
        return tgt

    def forward(
        self,
        tgt,
        memory,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        pos: Optional[Tensor] = None,
        query_pos: Optional[Tensor] = None,
    ):
        if self.normalize_before:
            return self.forward_pre(
                tgt,
                memory,
                tgt_mask,
                memory_mask,
                tgt_key_padding_mask,
                memory_key_padding_mask,
                pos,
                query_pos,
            )
        return self.forward_post(
            tgt,
            memory,
            tgt_mask,
            memory_mask,
            tgt_key_padding_mask,
            memory_key_padding_mask,
            pos,
            query_pos,
        )


def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(f"activation should be relu/gelu, not {activation}.")



# pylint: disable=all
# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Bowen Cheng from: https://github.com/facebookresearch/detr/blob/master/models/detr.py
import torch
from torch import nn
from torch.nn import functional as F

from ...weight_init import c2_xavier_fill
from .position_encoding import PositionEmbeddingSine
from .transformer import Transformer
from .mask2former_transformer_decoder import MultiScaleMaskedTransformerDecoder


def build_transformer_decoder(cfg, in_channels, mask_classification=True):
    """
    Build a instance embedding branch from `cfg.MODEL.INS_EMBED_HEAD.NAME`.
    """
    assert (name := cfg.MODEL.MASK_FORMER.TRANSFORMER_DECODER_NAME) == "MultiScaleMaskedTransformerDecoder", name
    params = MultiScaleMaskedTransformerDecoder.from_config(cfg, in_channels, mask_classification)
    return MultiScaleMaskedTransformerDecoder(**params)

class StandardTransformerDecoder(nn.Module):
    def __init__(
        self,
        in_channels,
        mask_classification=True,
        *,
        num_classes: int,
        hidden_dim: int,
        num_queries: int,
        nheads: int,
        dropout: float,
        dim_feedforward: int,
        enc_layers: int,
        dec_layers: int,
        pre_norm: bool,
        deep_supervision: bool,
        mask_dim: int,
        enforce_input_project: bool,
    ):
        """
        NOTE: this interface is experimental.
        Args:
            in_channels: channels of the input features
            mask_classification: whether to add mask classifier or not
            num_classes: number of classes
            hidden_dim: Transformer feature dimension
            num_queries: number of queries
            nheads: number of heads
            dropout: dropout in Transformer
            dim_feedforward: feature dimension in feedforward network
            enc_layers: number of Transformer encoder layers
            dec_layers: number of Transformer decoder layers
            pre_norm: whether to use pre-LayerNorm or not
            deep_supervision: whether to add supervision to every decoder layers
            mask_dim: mask feature dimension
            enforce_input_project: add input project 1x1 conv even if input
                channels and hidden dim is identical
        """
        super().__init__()

        self.mask_classification = mask_classification

        # positional encoding
        N_steps = hidden_dim // 2
        self.pe_layer = PositionEmbeddingSine(N_steps, normalize=True)

        transformer = Transformer(
            d_model=hidden_dim,
            dropout=dropout,
            nhead=nheads,
            dim_feedforward=dim_feedforward,
            num_encoder_layers=enc_layers,
            num_decoder_layers=dec_layers,
            normalize_before=pre_norm,
            return_intermediate_dec=deep_supervision,
        )

        self.num_queries = num_queries
        self.transformer = transformer
        hidden_dim = transformer.d_model

        self.query_embed = nn.Embedding(num_queries, hidden_dim)

        if in_channels != hidden_dim or enforce_input_project:
            self.input_proj = nn.Conv2d(in_channels, hidden_dim, kernel_size=1)
            c2_xavier_fill(self.input_proj)
        else:
            self.input_proj = nn.Sequential()
        self.aux_loss = deep_supervision

        # output FFNs
        if self.mask_classification:
            self.class_embed = nn.Linear(hidden_dim, num_classes + 1)
        self.mask_embed = MLP(hidden_dim, hidden_dim, mask_dim, 3)

    @classmethod
    def from_config(cls, cfg, in_channels, mask_classification):
        ret = {}
        ret["in_channels"] = in_channels
        ret["mask_classification"] = mask_classification

        ret["num_classes"] = cfg.MODEL.SEM_SEG_HEAD.NUM_CLASSES
        ret["hidden_dim"] = cfg.MODEL.MASK_FORMER.HIDDEN_DIM
        ret["num_queries"] = cfg.MODEL.MASK_FORMER.NUM_OBJECT_QUERIES
        # Transformer parameters:
        ret["nheads"] = cfg.MODEL.MASK_FORMER.NHEADS
        ret["dropout"] = cfg.MODEL.MASK_FORMER.DROPOUT
        ret["dim_feedforward"] = cfg.MODEL.MASK_FORMER.DIM_FEEDFORWARD
        ret["enc_layers"] = cfg.MODEL.MASK_FORMER.ENC_LAYERS
        ret["dec_layers"] = cfg.MODEL.MASK_FORMER.DEC_LAYERS
        ret["pre_norm"] = cfg.MODEL.MASK_FORMER.PRE_NORM
        ret["deep_supervision"] = cfg.MODEL.MASK_FORMER.DEEP_SUPERVISION
        ret["enforce_input_project"] = cfg.MODEL.MASK_FORMER.ENFORCE_INPUT_PROJ

        ret["mask_dim"] = cfg.MODEL.SEM_SEG_HEAD.MASK_DIM

        return ret

    def forward(self, x, mask_features, mask=None):
        if mask is not None:
            mask = F.interpolate(mask[None].float(), size=x.shape[-2:]).to(torch.bool)[0]
        pos = self.pe_layer(x, mask)

        src = x
        hs, memory = self.transformer(self.input_proj(src), mask, self.query_embed.weight, pos)

        if self.mask_classification:
            outputs_class = self.class_embed(hs)
            out = {"pred_logits": outputs_class[-1]}
        else:
            out = {}

        if self.aux_loss:
            # [l, bs, queries, embed]
            mask_embed = self.mask_embed(hs)
            outputs_seg_masks = torch.einsum("lbqc,bchw->lbqhw", mask_embed, mask_features)
            out["pred_masks"] = outputs_seg_masks[-1]
            out["aux_outputs"] = self._set_aux_loss(
                outputs_class if self.mask_classification else None, outputs_seg_masks
            )
        else:
            # FIXME h_boxes takes the last one computed, keep this in mind
            # [bs, queries, embed]
            mask_embed = self.mask_embed(hs[-1])
            outputs_seg_masks = torch.einsum("bqc,bchw->bqhw", mask_embed, mask_features)
            out["pred_masks"] = outputs_seg_masks
        return out

    @torch.jit.unused
    def _set_aux_loss(self, outputs_class, outputs_seg_masks):
        # this is a workaround to make torchscript happy, as torchscript
        # doesn't support dictionary with non-homogeneous values, such
        # as a dict having both a Tensor and a list.
        if self.mask_classification:
            return [
                {"pred_logits": a, "pred_masks": b}
                for a, b in zip(outputs_class[:-1], outputs_seg_masks[:-1])
            ]
        else:
            return [{"pred_masks": b} for b in outputs_seg_masks[:-1]]


class MLP(nn.Module):
    """Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(
            nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim])
        )

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x



"""semantic_representation.py -- helper class for all semantic segmentation representations"""
from overrides import overrides
import numpy as np
from vre.representations import Representation, NpIORepresentation
from vre.utils import DiskData, MemoryData, image_resize_batch, colorize_semantic_segmentation, ReprOut
from vre.logger import vre_logger as logger

class SemanticRepresentation(Representation, NpIORepresentation):
    """SemanticRepresentation. Implements semantic task-specific stuff, like argmaxing if needed"""
    def __init__(self, *args, classes: int | list[str], color_map: list[tuple[int, int, int]],
                 disk_data_argmax: bool, **kwargs):
        self.n_classes = len(list(range(classes)) if isinstance(classes, int) else classes)
        Representation.__init__(self, *args, **kwargs)
        NpIORepresentation.__init__(self)
        self.classes = list(range(classes)) if isinstance(classes, int) else classes
        self.color_map = color_map
        self.disk_data_argmax = disk_data_argmax
        assert len(color_map) == self.n_classes and self.n_classes > 1, (self.name, color_map, self.n_classes)
        self._output_dtype = "uint8" if disk_data_argmax else "float16"

    @property
    @overrides
    def n_channels(self) -> int:
        return self.n_classes

    @overrides
    def disk_to_memory_fmt(self, disk_data: DiskData) -> MemoryData:
        memory_data = MemoryData(disk_data)
        if self.disk_data_argmax:
            if disk_data.dtype not in (np.uint8, np.uint16):
                # below case is for distillation where in some cases we need float32 (perhaps refactor someplace else?)
                logger.debug2(f"[self.name]. {disk_data.dtype=} but self.disk_data_argmax is True")
                memory_data = MemoryData(disk_data)
            else:
                memory_data = MemoryData(np.eye(len(self.classes))[disk_data].astype(np.float32))
        assert memory_data.dtype == np.float32 and memory_data.shape[-1] == self.n_classes, (self.name, memory_data)
        return memory_data

    @overrides
    def memory_to_disk_fmt(self, memory_data: MemoryData) -> DiskData:
        assert memory_data.shape[-1] == self.n_classes, (self.name, memory_data.shape, self.n_classes)
        return memory_data.argmax(-1) if self.disk_data_argmax else memory_data

    @overrides
    def make_images(self, data: ReprOut) -> np.ndarray:
        assert data is not None, f"[{self}] data must be first computed using compute()"
        frames_rsz = None
        if data.frames is not None:
            frames_rsz = image_resize_batch(data.frames, *data.output.shape[1:3])
        return colorize_semantic_segmentation(data.output.argmax(-1), self.classes, self.color_map, rgb=frames_rsz)



"""SafeUAV semanetic segmentation representation"""
import sys
from pathlib import Path
from overrides import overrides
import numpy as np
import torch as tr
from torch.nn import functional as F

from vre.vre_video import VREVideo
from vre.logger import vre_logger as logger
from vre.utils import MemoryData, image_read, image_write
from vre.representations import ReprOut, LearnedRepresentationMixin, ComputeRepresentationMixin
from vre_repository.weights_repository import fetch_weights
from vre_repository.semantic_segmentation import SemanticRepresentation

try:
    from .model import SafeUAV as Model
except ImportError:
    from model import SafeUAV as Model

class SafeUAV(SemanticRepresentation, LearnedRepresentationMixin, ComputeRepresentationMixin):
    """SafeUAV semantic segmentation representation"""
    def __init__(self, disk_data_argmax: bool, variant: str, **kwargs):
        LearnedRepresentationMixin.__init__(self)
        ComputeRepresentationMixin.__init__(self)
        self.variant = variant
        assert variant in ("model_1M", "model_4M", "model_430k", "testing"), variant
        color_map = [[0, 255, 0], [0, 127, 0], [255, 255, 0], [255, 255, 255],
                     [255, 0, 0], [0, 0, 255], [0, 255, 255], [127, 127, 63]]
        classes = ["land", "forest", "residential", "road", "little-objects", "water", "sky", "hill"]
        SemanticRepresentation.__init__(self, classes=classes, color_map=color_map,
                                        disk_data_argmax=disk_data_argmax, **kwargs)
        self.model: Model | None = None
        self.cfg: dict | None = None
        self.statistics: dict[str, list[float]] | None = None
        self.output_dtype = "uint8" if disk_data_argmax else "float16"

    @property
    @overrides
    def n_channels(self) -> int:
        return len(self.classes)

    @overrides
    def compute(self, video: VREVideo, ixs: list[int]):
        assert self.data is None, f"[{self}] data must not be computed before calling this"
        h, w = self.cfg["model"]["hparams"]["data_shape"]["rgb"][1:3]
        cumsum = [0, *np.cumsum([x[0] for x in self.cfg["model"]["hparams"]["data_shape"].values()])]
        rgb_pos = self.cfg["data"]["parameters"]["task_names"].index("rgb")
        x = tr.zeros(len(ixs), self.model.encoder.d_in, h, w, device=self.device)
        tr_rgb = F.interpolate(tr.from_numpy(video[ixs]).permute(0, 3, 1, 2).to(self.device), size=(h, w))
        mean, std = self.statistics["rgb"][2:4]
        tr_rgb = (tr_rgb - tr.Tensor(mean).reshape(1, 3, 1, 1)) / tr.Tensor(std).reshape(1, 3, 1, 1)
        x[:, cumsum[rgb_pos]: cumsum[rgb_pos+1] ] = tr_rgb

        with tr.no_grad():
            y = self.model.forward(x)
        sema_pos = self.cfg["data"]["parameters"]["task_names"].index("semantic_output")
        y_rgb = y[:, cumsum[sema_pos]: cumsum[sema_pos+1]].permute(0, 2, 3, 1).cpu().numpy()
        self.data = ReprOut(frames=video[ixs], output=MemoryData(y_rgb), key=ixs)

    @staticmethod
    @overrides
    def weights_repository_links(**kwargs: dict) -> list[str]:
        assert (variant := kwargs["variant"]) != "testing", variant
        return [f"semantic_segmentation/safeuav/{variant}.ckpt"]

    @overrides
    def vre_setup(self, load_weights: bool = True):
        assert self.setup_called is False
        if self.variant == "testing":
            self.model = Model(in_channels=15, out_channels=15, num_filters=8)
            self.cfg = {
                "model": {
                    "hparams": {
                        "data_shape": {
                            "camera_normals_output": [3, 50, 50], "depth_output": [1, 50, 50],
                            "rgb": [3, 50, 50], "semantic_output": [8, 50, 50]
                        },
                    }
                },
                "data": {
                    "parameters": {
                        "task_names": ['depth_output', 'camera_normals_output', 'rgb', 'semantic_output']
                    }
                }
            }
            self.statistics = {
                "rgb": [[0.0, 0.0, 0.0],
                        [255.0, 255.0, 255.0],
                        [95.4980582891499, 103.0128696717979, 85.04794782640131],
                        [55.569107621001216, 58.133399225741776, 61.12564370234619]],
            }
        else:
            assert load_weights is True, load_weights
            ckpt = tr.load(fetch_weights(SafeUAV.weights_repository_links(variant=self.variant))[0], map_location="cpu")
            self.cfg = ckpt["hyper_parameters"]["cfg"]
            self.statistics = ckpt["hyper_parameters"]["statistics"]
            self.model = Model(**self.cfg["model"]["parameters"])
            self.model.load_state_dict(ckpt["state_dict"])

        self.model = self.model.eval().to(self.device)
        self.setup_called = True

    @overrides
    def vre_free(self):
        assert self.setup_called is True and self.model is not None, (self.setup_called, self.model is not None)
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None
        self.cfg = None
        self.statistics = None
        self.setup_called = False

if __name__ == "__main__":
    img = image_read(sys.argv[1])
    assert len(sys.argv) == 3, "Usage: python safeuav.py /path/to/img.png [model_430k/model_1M/model_4M]"
    model = SafeUAV(name="safeuav", disk_data_argmax=True, variant=sys.argv[2])
    model.vre_setup(load_weights=True)

    model.compute(img[None], [0])
    res_img = model.make_images(model.data)[0]
    out_path = Path(__file__).parent / f"{Path(sys.argv[1]).stem}.png"
    image_write(res_img, out_path)
    logger.info(f"Stored prediction at '{out_path}'")



"""RGB representation."""
from __future__ import annotations
from overrides import overrides

from vre.utils import MemoryData, ReprOut
from vre.vre_video import VREVideo
from vre.representations import ComputeRepresentationMixin

from .color_representation import ColorRepresentation

class RGB(ColorRepresentation, ComputeRepresentationMixin):
    """Basic RGB representation."""
    def __init__(self, *args, **kwargs):
        ColorRepresentation.__init__(self, *args, **kwargs)
        ComputeRepresentationMixin.__init__(self)
        assert len(self.dependencies) == 0, self.dependencies
        self.output_dtype = "uint8"

    @overrides
    def compute(self, video: VREVideo, ixs: list[int]):
        assert self.data is None, f"[{self}] data must not be computed before calling this"
        self.data = ReprOut(frames=video[ixs], output=MemoryData(video[ixs]), key=ixs) # video[ixs] is cached



"""IORepresentation: representations that are files on the disk and not computed from video or other representations"""
from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
import numpy as np

from vre.logger import vre_logger as logger
from vre.utils import image_resize_batch, MemoryData, DiskData
from .representation import ReprOut

class BinaryFormat(Enum):
    """types of binary outputs from a representation"""
    NOT_SET = "not-set"
    NPZ = "npz"
    NPY = "npy"

class ImageFormat(Enum):
    """types of image outputs from a representation"""
    NOT_SET = "not-set"
    PNG = "png"
    JPG = "jpg"

class IORepresentationMixin(ABC):
    """
    StoredRepresentation. These methods define the blueprint to store and load a representation from disk to RAM.
    Data transformations order:
    - DISK [.npz] -> load_from_disk() -> [disk_fmt] -> disk_to_memory_fmt() -> RAM [memory_fmt] (stored in .data)
    - RAM [memory_fmt] -> memory_to_disk_fmt() -> [disk_fmt] -> store_to_disk() -> DISK [.npz]
    The intermediate disk_fmt is only used for storage efficiency, while all the vre operations happen with memory_fmt
    For example: disk_fmt is a binary bool vector, while memory_fmt is a one-hot 2 channels map.
    """
    def __init__(self):
        self._binary_format: BinaryFormat | None = None
        self._image_format: ImageFormat | None = None
        self._compress: bool | None = None
        self._output_size: tuple[int, int] | str | None = None
        self._output_dtype: str | np.dtype | None = None

    @abstractmethod
    def load_from_disk(self, path: Path) -> DiskData:
        """Reads the data from the disk into disk_fmt"""

    @abstractmethod
    def disk_to_memory_fmt(self, disk_data: DiskData) -> MemoryData:
        """Transforms the data from disk_fmt into memory_fmt (usable in VRE)"""

    @abstractmethod
    def memory_to_disk_fmt(self, memory_data: MemoryData) -> DiskData:
        """Transformes the data from memory_fmt (usable in VRE) to disk_fmt"""

    @abstractmethod
    def save_to_disk(self, memory_data: MemoryData, path: Path):
        """Stores the disk_fmt data to disk"""

    @property
    def binary_format(self) -> BinaryFormat:
        """the binary format of the representation"""
        if self._binary_format is None:
            logger.warning(f"[{self}] No binary_format set, returning not-set. Call set_io_params")
            return BinaryFormat.NOT_SET
        return self._binary_format

    @binary_format.setter
    def binary_format(self, bf: BinaryFormat | str):
        assert isinstance(bf, (BinaryFormat, str)), f"Must be one of {[_bf.value for _bf in BinaryFormat]}. Got {bf}."
        bf = BinaryFormat(bf) if isinstance(bf, str) else bf
        self._binary_format = bf

    @property
    def export_binary(self) -> bool:
        """whether to export binary or not"""
        return self.binary_format != BinaryFormat.NOT_SET

    @property
    def image_format(self) -> ImageFormat:
        """the image format of the representation"""
        if self._image_format is None:
            logger.warning(f"[{self}] No image_format set, returning not-set. Call set_io_params")
            return ImageFormat.NOT_SET
        return self._image_format

    @image_format.setter
    def image_format(self, imf: ImageFormat | str):
        assert isinstance(imf, (ImageFormat, str)), f"Must be one of {[_imf.value for _imf in ImageFormat]}. Got {imf}."
        imf = ImageFormat(imf) if isinstance(imf, str) else imf
        self._image_format = imf

    @property
    def export_image(self) -> bool:
        """whether to export image or not"""
        return self.image_format != ImageFormat.NOT_SET

    @property
    def compress(self) -> bool:
        """whether to compress the output or not"""
        if self._compress is None:
            logger.warning(f"[{self}] No compress set, returning True. Call set_io_params")
            return True
        return self._compress

    @compress.setter
    def compress(self, c: bool):
        self._compress = c

    @property
    def output_size(self) -> str | tuple[int, int]:
        """Returns the output size as a tuple/string or None if it's not explicitly set"""
        if self._output_size is None:
            logger.warning(f"[{self}] No output_size set, returning 'video_shape'. Call set_compute_params")
            return "video_shape"
        return self._output_size

    @output_size.setter
    def output_size(self, os: str | tuple[int, int]):
        assert isinstance(os, (str, tuple, list)), os
        if isinstance(os, (tuple, list)):
            assert len(os) == 2 and all(isinstance(x, int) and x > 0 for x in os), os
        if isinstance(os, str):
            assert os in ("native", "video_shape"), os
        self._output_size = os if isinstance(os, str) else tuple(os)

    @property
    def output_dtype(self) -> np.dtype:
        """the output dtype of the representation"""
        if self._output_dtype is None:
            return np.float32
        return self._output_dtype

    @output_dtype.setter
    def output_dtype(self, dtype: str | np.dtype):
        assert isinstance(dtype, (str, np.dtype)), dtype
        self._output_dtype = np.dtype(dtype)

    def set_io_params(self, **kwargs):
        """set the IO parameters for the representation"""
        attributes = ["binary_format", "image_format", "compress", "output_size", "output_dtype"]
        res = ""
        assert set(kwargs).issubset(attributes), (list(kwargs), attributes)
        for attr in attributes:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
                res += f"\n-{attr}: {kwargs[attr]}"
        if len(res) > 0:
            logger.debug(f"[{self}] Set node specific 'IO' params: {res}")

    def resize(self, data: ReprOut, new_size: tuple[int, int]):
        """resizes the data. size is provided in (h, w)"""
        assert data is not None, "No data provided"
        interpolation = "nearest" if np.issubdtype(d := data.output.dtype, np.integer) or d == bool else "bilinear"
        output_images = None
        if data.output_images is not None:
            output_images = image_resize_batch(data.output_images, *new_size, interpolation="nearest")
        return ReprOut(frames=data.frames, key=data.key, extra=data.extra, output_images=output_images,
                       output=image_resize_batch(data.output, *new_size, interpolation=interpolation))
